"""Tests roles.py module."""
from typing import Type
from unittest.mock import Mock

import pytest
from pytest import fixture

from bitfount.federated.roles import Role, _RolesMixIn
from tests.utils.helper import unit_test


@unit_test
class TestRolesMixIn:
    """Test RolesMixIn class."""

    @fixture
    def dummy_role_based_class(self) -> Type[_RolesMixIn]:
        """Returns dummy class that subclasses RolesMixIn."""

        class DummyClass(_RolesMixIn):
            def modeller(self) -> Mock:
                return Mock(role="modeller")

            def worker(self) -> Mock:
                return Mock(role="worker")

        return DummyClass

    def test_roles(self, dummy_role_based_class: Type[_RolesMixIn]) -> None:
        """Check that supported roles property works as expected."""
        assert dummy_role_based_class().roles == {Role.MODELLER, Role.WORKER}

    @pytest.mark.parametrize("role", ["modeller", "worker", "non_existent_role"])
    def test_create_supported_role(
        self, dummy_role_based_class: Type[_RolesMixIn], role: str
    ) -> None:
        """Check we can create roles properly by name."""
        available_roles = [item.value for item in Role]
        obj = dummy_role_based_class()
        if role in available_roles:
            modeller = obj.create(role)
            assert modeller.role == role
        else:
            with pytest.raises(ValueError):
                obj.create(role)



"""Algorithm to train a model remotely and return its parameters."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, Mapping, Optional, Tuple

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithm,
    _BaseModelAlgorithmFactory,
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.hub.api import BitfountHub
from bitfount.models.base_models import MAIN_MODEL_REGISTRY
from bitfount.types import (
    T_NESTED_FIELDS,
    DistributedModelProtocol,
    _Residuals,
    _SerializedWeights,
    _Weights,
)
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.types import _DistributedModelTypeOrReference

logger = _get_federated_logger(__name__)


class _BaseModelTrainingMixIn(_BaseModelAlgorithm):
    """Shared methods/attributes for both modeller and worker."""

    # This is set in the base model algorithm
    model: DistributedModelProtocol

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    @property
    def epochs(self) -> Optional[int]:
        """Returns model epochs."""
        return self.model.epochs

    @property
    def steps(self) -> Optional[int]:
        """Returns model steps."""
        return self.model.steps

    def diff_params(self, old_params: _Weights) -> _Residuals:
        return self.model.diff_params(
            old_params=old_params, new_params=self.model.get_param_states()
        )

    def get_param_states(self) -> _Weights:
        """Returns the current parameters of the underlying model."""
        return self.model.get_param_states()

    def apply_update(self, update: _Weights) -> _Weights:
        """Applies a parameter update to the underlying model."""
        return self.model.apply_weight_updates([update])

    def update_params(self, params: _Weights) -> None:
        """Updates model parameters."""
        return self.model.update_params(params)


class _ModellerSide(
    _BaseModelTrainingMixIn,
    _BaseModellerModelAlgorithm,
):
    """Modeller side of the FederatedModelTraining algorithm."""

    def __init__(
        self,
        *,
        model: DistributedModelProtocol,
        **kwargs: Any,
    ):
        super().__init__(model=model, **kwargs)

    def run(
        self,
        update: Optional[_Weights] = None,
        validation_metrics: Optional[Mapping[str, float]] = None,
    ) -> _SerializedWeights:
        """Takes a weight update, applies it and returns the new model parameters."""
        if update is not None:
            self.apply_update(update)
        nn_params: _Weights = self.get_param_states()
        serialized_params = self.model.serialize_params(nn_params)
        if validation_metrics:
            for key, value in validation_metrics.items():
                self.model.log_(key, value, on_epoch=True, prog_bar=True, logger=True)
        return serialized_params


class _WorkerSide(
    _BaseModelTrainingMixIn,
    _BaseWorkerModelAlgorithm,
):
    """Worker side of the FederatedModelTraining algorithm."""

    def __init__(
        self,
        *,
        model: DistributedModelProtocol,
        **kwargs: Any,
    ):
        super().__init__(model=model, **kwargs)

    def run(
        self,
        data: BaseSource,
        serialized_model_params: _SerializedWeights,
        iterations: int,
    ) -> Tuple[_Residuals, Optional[Dict[str, str]]]:
        """Takes the model parameters, trains and returns the parameter update."""
        model_params = self.model.deserialize_params(serialized_model_params)
        self.update_params(model_params)

        # Train for one federated round - `iterations` many steps or epochs
        self.model.set_model_training_iterations(iterations)
        self.model.reset_trainer()
        validation_metrics: Optional[Dict[str, str]] = self.model.fit(data)
        # Return the weight update and validation metrics
        return self.diff_params(old_params=model_params), validation_metrics

    def save_final_parameters(self, model_params: _SerializedWeights) -> None:
        """Saves the final global model parameters.

        :::note

        This method saves the final global model to a file called `model.pt`.

        :::

        Args:
            model_params: The final global model parameters.
        """
        tensor_model_params = self.model.deserialize_params(model_params)
        self.update_params(tensor_model_params)
        # TODO: [BIT-1043]: pass filename for serialization
        self.model.serialize("model.pt")


@delegates()
class FederatedModelTraining(
    _BaseModelAlgorithmFactory,
):
    """Algorithm for training a model remotely and returning its updated parameters.

    This algorithm is designed to be compatible with the `FederatedAveraging` protocol.

    Args:
        model: The model to train on remote data.

    Attributes:
        model: The model to train on remote data.
    """

    nested_fields: ClassVar[T_NESTED_FIELDS] = {"model": MAIN_MODEL_REGISTRY}

    def __init__(
        self,
        *,
        model: _DistributedModelTypeOrReference,
        **kwargs: Any,
    ):
        super().__init__(model=model, **kwargs)

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the FederatedModelTraining algorithm."""
        model = self._get_model_from_reference()
        return _ModellerSide(model=model, **kwargs)

    def worker(self, hub: BitfountHub, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the FederatedModelTraining algorithm.

        Args:
            hub: `BitfountHub` object to use for communication with the hub.
        """
        model = self._get_model_from_reference(hub=hub)
        return _WorkerSide(model=model, **kwargs)



"""Helper functions related to hub and AM interactions."""
import logging
from pathlib import Path
from typing import Dict, Iterable, Mapping, Optional, Union

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
import yaml

from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    BITFOUNT_KEY_STORE,
    _get_environment,
)
from bitfount.data.schema import BitfountSchema
from bitfount.federated.encryption import _RSAEncryption
from bitfount.hub.api import BitfountAM, BitfountHub
from bitfount.hub.authentication_flow import (
    _DEFAULT_USERNAME,
    _DEVELOPMENT_AUTH_DOMAIN,
    _DEVELOPMENT_CLIENT_ID,
    _STAGING_AUTH_DOMAIN,
    _STAGING_CLIENT_ID,
    BitfountSession,
)
from bitfount.hub.exceptions import PodDoesNotExistError
from bitfount.hub.types import (
    _DEV_AM_URL,
    _DEV_HUB_URL,
    _STAGING_AM_URL,
    _STAGING_HUB_URL,
    PRODUCTION_AM_URL,
    PRODUCTION_HUB_URL,
)

logger = logging.getLogger(__name__)


def _create_bitfount_session(
    url: str, username: Optional[str] = None
) -> BitfountSession:
    """Creates a relevant Bitfount Session according to the environment.

    Args:
        url: Bitfount hub URL.
        username: Optional. Username. Defaults to DEFAULT_USER_DIRECTORY.

    Returns:
        BitfountSession object.
    """
    username = username if username else _DEFAULT_USERNAME

    if url == _STAGING_HUB_URL:
        session = BitfountSession(
            auth_domain=_STAGING_AUTH_DOMAIN,
            client_id=_STAGING_CLIENT_ID,
            username=username,
        )
    elif "localhost" in url:
        session = BitfountSession(
            auth_domain=_DEVELOPMENT_AUTH_DOMAIN,
            client_id=_DEVELOPMENT_CLIENT_ID,
            username=username,
        )
    else:
        session = BitfountSession(username=username)

    return session


def _default_bitfounthub(
    hub: Optional[BitfountHub] = None,
    username: Optional[str] = None,
    url: Optional[str] = None,
) -> BitfountHub:
    """Gets a default BitfountHub instance if one is not specified.

    Args:
        hub: Optional. The BitfountHub instance to use if it exists.
        username: Optional. Username.
        url: Optional. Bitfount hub URL.

    Returns:
        BitfountHub object representing the hub.
    """
    if not hub:
        return _create_bitfounthub(username, url)
    return hub


def _create_bitfounthub(
    username: Optional[str] = None, url: Optional[str] = None
) -> BitfountHub:
    """Creates bitfounthub object.

    Args:
        username: Optional. Username.
        url: Optional. Bitfount hub URL. Will use the environment default if not
             supplied.

    Returns:
        BitfountHub object representing the hub.
    """
    if not url:
        environment = _get_environment()
        if environment == _STAGING_ENVIRONMENT:
            url = _STAGING_HUB_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            url = _DEV_HUB_URL
        else:
            url = PRODUCTION_HUB_URL
    session = _create_bitfount_session(url=url, username=username)
    session.authenticate()
    return BitfountHub(session=session, url=url)


def _create_access_manager(
    session: BitfountSession, url: Optional[str] = None
) -> BitfountAM:
    """Creates and returns Bitfount Access Manager.

    Args:
        session: Bitfount session for authentication.
        url: Optional. URL of the access manager. Will use the environment default
             if not supplied.

    Returns:
        The BitfountAM representing the access manager.
    """
    if not url:
        environment = _get_environment()
        if environment == _STAGING_ENVIRONMENT:
            url = _STAGING_AM_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            url = _DEV_AM_URL
        else:
            url = PRODUCTION_AM_URL

    return BitfountAM(session, url)


def _save_key_to_key_store(
    key_store_path: Path, pod_identifier: str, serialized_pod_key: str
) -> None:
    """Save pod keys to the key store.

    This will override any previously saved keys.

    Args:
        key_store_path: The path to where the key files are stored
        pod_identifier: The pod identifier of the pod we are saving the key for.
        serialized_pod_key: The serialized pod key
    """
    with open(key_store_path, "r") as key_store:
        known_pod_keys: Dict[str, str] = yaml.safe_load(key_store) or {}

    known_pod_keys[pod_identifier] = serialized_pod_key

    with open(key_store_path, "w") as key_store:
        logger.debug(f"Saving pod public keys to {key_store_path}")
        yaml.safe_dump(known_pod_keys, key_store)


def _check_known_pods(
    pod_name: str, pod_public_key: RSAPublicKey, key_store_path: Path
) -> RSAPublicKey:
    """Checks known pods still have valid public key.

    Checks pod public key is the same as in BITFOUNT_KEY_STORE
    If BITFOUNT_KEY_STORE does not exist, it creates it and adds the key
    If it does exist but the key is different, it prompts the user to
    accept or reject new key.

    Returns:
        The new or existing key (depending on which the user chooses) as a string.
    """
    # Serialise to allow comparison with stored keys
    serialized_pod_key: str = _RSAEncryption.serialize_public_key(
        pod_public_key
    ).decode()

    # Check if target key already exists in the key store
    with open(key_store_path, "r") as key_store:
        known_pod_keys: Dict[str, str] = yaml.safe_load(key_store) or {}
        serialized_current_pod_key: Optional[str] = known_pod_keys.get(pod_name, None)

        if serialized_current_pod_key is None:
            # No current key for this user, ok to "update" it (i.e. add it)
            logger.debug(f"No existing key found for pod {pod_name}")
        elif serialized_current_pod_key != serialized_pod_key:
            # Key mismatch, check with user for how to proceed
            logger.error(
                f"{pod_name} public key has changed. \n"
                f"Please double check the key has really changed and this "
                f"is not an attack."
            )
            while True:
                # We are only using python3 so this is secure
                response = input(  # nosec
                    "Do you want to trust the new public key? (y/n)"
                )
                if response.lower() == "y":
                    # Break out of the loop
                    break
                elif response.lower() == "n":
                    # Use the old public key
                    return _RSAEncryption.load_public_key(
                        serialized_current_pod_key.encode()
                    )
                else:
                    print("Didn't catch that, please type 'y' or 'n'")
        else:
            # Key matches, just return it
            logger.info(f"Found public key for {pod_name} in key store.")
            return pod_public_key

    # Save known pod key with the new one. We do this here because either
    # the user has approved or there was no key for that user in the first
    # place.
    logger.info(f"Saving public key for {pod_name} in key store.")
    _save_key_to_key_store(key_store_path, pod_name, serialized_pod_key)

    # The new pod key has been saved so we can just return it
    return pod_public_key


def _get_pod_public_key(
    pod_identifier: str,
    hub: BitfountHub,
    pod_public_key_paths: Optional[Mapping[str, Path]] = None,
) -> RSAPublicKey:
    """Gets a pod's public key.

    Either loads it from disk or downloads it from BitfountHub.

    Args:
        pod_identifier: The pod identifier of the pod we get the public key for.
        hub: Hub to download keys from.
        pod_public_key_paths: Optional. Mapping of pod identifiers to already
            existing key files.

    Returns:
        The public key for the pod.

    Raises:
        PodDoesNotExistError: If the pod does not exist.
    """
    key_store_path = Path(BITFOUNT_KEY_STORE).expanduser()

    # Create key store file if file doesn't exist
    if not key_store_path.is_file():
        logger.info("Creating key store for pod public keys...")
        key_store_path.parent.mkdir(parents=True, exist_ok=True)
        key_store_path.touch()

    # Check for the target key against explicitly provided key files
    if pod_public_key_paths:
        logger.debug(f"Checking for public key file for {pod_identifier}")
        try:
            pod_public_key_path: Path = pod_public_key_paths[pod_identifier]
            existing_pod_public_key = _RSAEncryption.load_public_key(
                pod_public_key_path
            )
            return _check_known_pods(
                pod_identifier, existing_pod_public_key, key_store_path
            )
        except KeyError:
            # We have some pod public key files, but not for this pod identifier
            logger.debug(f"No existing public key file for {pod_identifier}")

    # If no key files at all, or just not for that pod, retrieve key
    if pod_key := hub.get_pod_key(pod_identifier=pod_identifier):
        pod_public_key = _RSAEncryption.load_public_key(pod_key.encode())

        # Check that the downloaded key matches any existing one we have and save
        # it to the key store.
        return _check_known_pods(pod_identifier, pod_public_key, key_store_path)
    else:
        raise PodDoesNotExistError(f"No public key found for pod: {pod_identifier}")


def _get_pod_public_keys(
    pod_identifiers: Iterable[str],
    hub: BitfountHub,
    pod_public_key_paths: Optional[Mapping[str, Path]] = None,
) -> Dict[str, RSAPublicKey]:
    """Retrieve the public keys for a group of pods.

    Either loads it from disk or downloads it from BitfountHub.

    Args:
        pod_identifiers: The pod identifiers of the pods we want the public keys for.
        hub: Hub to download keys from.
        pod_public_key_paths: Optional. Mapping of pod identifiers to already
                              existing keys.

    Returns:
        A dictionary of pod identifiers to the public key for that pod.
    """
    # Either download public key, or load existing, for each pod
    loaded_pod_public_keys: Dict[str, RSAPublicKey] = {
        pod_identifier: _get_pod_public_key(pod_identifier, hub, pod_public_key_paths)
        for pod_identifier in pod_identifiers
    }
    return loaded_pod_public_keys


def get_pod_schema(
    pod_identifier: str,
    save_file_path: Optional[Union[str, Path]] = None,
    hub: Optional[BitfountHub] = None,
    username: Optional[str] = None,
) -> BitfountSchema:
    """Get a pod's schema from the hub.

    Args:
        pod_identifier: The identifier of the pod. If supplied with only pod name
            assumes the namespace is the current user.
        save_file_path: Optional. Path to save the downloaded schema to. Won't save
            if not provided.
        hub: Optional. The BitfountHub to connect to. The default hub will be
             used if not provided.
        username: The username to use when connecting to the hub if a hub instance is
            not provided.

    Returns:
        The loaded BitfountSchema object.
    """
    # Generate default hub if not provided
    if not hub:
        hub = _create_bitfounthub(username=username)
    elif username:
        logger.warning("Ignoring username argument as hub was provided.")

    # Check if full pod_identifier or pod name only
    if "/" not in pod_identifier:
        # Construct full pod identifier if needed
        pod_identifier = f"{hub.username}/{pod_identifier}"

    schema = hub.get_pod_schema(pod_identifier)

    # Save out schema if requested
    if save_file_path:
        schema.dump(Path(save_file_path).expanduser())
    return schema



"""Testing helper functions."""
import datetime
import inspect
from io import BytesIO
import logging
from pathlib import Path
import random
import string
from typing import Any, List, Literal, Optional, Protocol, Sequence, Tuple, Union, cast
from unittest.mock import create_autospec

from PIL import Image, ImageOps
from _pytest.logging import LogCaptureFixture
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
import numpy as np
import pandas as pd
import pytest

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import SchemaOverrideMapping
from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.base import _BaseAggregatorFactory
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.authorisation_checkers import (
    IdentityVerificationMethod,
    _LocalAuthorisation,
)
from bitfount.federated.early_stopping import FederatedEarlyStopping
from bitfount.federated.modeller import _Modeller
from bitfount.federated.pod_response_message import _PodResponseMessage
from bitfount.federated.secure import SecureShare
from bitfount.federated.transport.worker_transport import (
    _InterPodWorkerMailbox,
    _WorkerMailbox,
)
from bitfount.federated.types import AlgorithmType, ProtocolType
from bitfount.federated.utils import _ALGORITHMS, _MODELS, _PROTOCOLS
from bitfount.federated.worker import _Worker
from bitfount.hub.api import BitfountHub
from bitfount.metrics import (
    BINARY_CLASSIFICATION_METRICS,
    MULTICLASS_CLASSIFICATION_METRICS,
    REGRESSION_METRICS,
    MetricCollection,
)
from bitfount.models.base_models import ClassifierMixIn, Optimizer, _BaseModel
from bitfount.types import DistributedModelProtocol, _StrAnyDict
from bitfount.utils import ExampleSegmentationData, _add_this_to_list, seed_all
from tests.utils.mocks import LocalMessageService, LocalMessageServiceSharedQueues

# PYTEST MARKS
# tests involving a specific backend
backend_test = pytest.mark.backend_test

# tests that run the full end-to-end system in the form of tutorials
tutorial_test = pytest.mark.tutorial_test
# tests that run the full end-to-end system, but mock out web calls
end_to_end_mocks_test = pytest.mark.end_to_end_mocks_test
# tests that run the full end-to-end system, including web calls
end_to_end_test = pytest.mark.end_to_end_test
# slower, integration- or training-like tests
integration_test = pytest.mark.integration_test
# faster, unit-test-like tests
unit_test = pytest.mark.unit_test
# Individual test markers for tutorial tests
tutorial_1_test = pytest.mark.tutorial_1_test
tutorial_2_test = pytest.mark.tutorial_2_test
tutorial_3_test = pytest.mark.tutorial_3_test
tutorial_4_test = pytest.mark.tutorial_4_test
tutorial_5_test = pytest.mark.tutorial_5_test
tutorial_6_test = pytest.mark.tutorial_6_test
tutorial_7_test = pytest.mark.tutorial_7_test
tutorial_8_test = pytest.mark.tutorial_8_test
tutorial_9_test = pytest.mark.tutorial_9_test
tutorial_10_test = pytest.mark.tutorial_10_test
tutorial_11_test = pytest.mark.tutorial_11_test

DATASET_ROW_COUNT = 4000
DIMS = (DATASET_ROW_COUNT, 4)
AUC_THRESHOLD = 0.6
MAE_THRESHOLD = 0.2

PRIVATE_KEY = inspect.cleandoc(
    """-----BEGIN RSA PRIVATE KEY-----
    MIIEpQIBAAKCAQEAwA0cNMjzEm+LPFolbdOAcIB6hX7QGqkzVC8L8W5X2qFFDrSb
    TI9CSifQ3/1A+xYfiZrR4Za2cCvCtIzHxVctEAcQSW+IDRjh5vvdGTKgJMtaf3YH
    eXMk9NAN4Px6McuqaBE2llYt6okICha8MEUR9lffa761WkOOWnWAXQFUWxWOxuVn
    KQOOm1OQ0OiGd+6b81TbSmHDL1QeAMN66tscnR0zFuLfaCRz8yZIsKR1bUPTAUnC
    B3gaUXIixxDTeSG6XHItb7c/uHJUPtW4lFLiDF12JX0GwQvkpYCe3I++fBxC0xcd
    GwmncKocOLtsuOPswDHH3waxp5KGVDsdx8Po8QIDAQABAoIBAQCiYvN0lm0hK3u5
    nieDw5AYkCoI8n8X6/Vgux8IF1rlm/L+Siv+Aiv089Gjc0ochEpKWcQZUQwVsIrD
    iz3nWJzbrn1Q+qM5H87zNgdvJOVP9zogCDattHSNI9Z4ZEMWE7WUpOlZGKleZFuN
    3EdnMR/NAeQCAJDrL7AQQMY23cbr88dZ1VJ0zJdt50IdjTHjG5gpX8ifl3egWQSa
    xKh3QJWFQI8oR69lT29ZAGgaPb3mhruyvUCK0ioI++g9kBMVncMeNV0SrsHNTdWH
    mKhjEI1KeNTPBXxZCD43L2/tCp+K+9WZs2LJdsCtN1/bDjc7xMWFQnoJ+8OcpYEX
    tV5k+VgBAoGBAPapKUk7vtaLhCA2+T7dde83QjgvcNwH2hB7Zbs5rEElBJUWDFsf
    Av08jqHY8kTsBbh78uq+CWTY+G/d2pUw6hHoHVWLyTrMbnIl/moRjypCNsKcSRP/
    tmAepWz3smBfpgin5GrSx29LCI6EE1sOOKYj1IsQPwuTObFqnCBa8e4xAoGBAMdS
    oMtTI4eNR7X8jS+id0rM3tvWwxQT5SvyoJA/WzQa1PRCf96r6CXWZ7+0A38tpkdF
    OL3qK2UUzweVqEGZu+ZzrrOBO9d2jfbcZSSZ/Qd3fANpl9SkImYVyyZhGzbkIZce
    5UkH3HauLDvvCnWl2JJDlzgJuqQu+F9AKmxj8TbBAoGBANYHlOaR7Bc4Om7FpOLy
    wZnQBirAp5qVSZfRq0WZVqm0WFSRLCmJMEvS1yUDHb3GW3RoQTGFspsKVhCjnj6k
    kgb1tTZ8tCDMMO3ln+wjzmlpySB3AkZTNcoT7ZEvIV3UwTpzjJVQznL41bTsT+tL
    3MX1A0Gj2EDQAFi9XDoLEDCRAoGBAI69sKMLTfVkCsfcLevACHgL0UlUOm3ldYgM
    gEns5U6iLIEkIlh/eYenTLMvohVwxpRhkSoISb1gsoZ2+YDJLJPzolewo/9ASphP
    6yrUEC1JtwjmlLHWPUAoOUyp3AhqRVfo60BqncpsHwdov2/TLg5I00FxTjUH4hys
    YfUveW3BAoGAEjzWgY8QAURahV9yM+OHmeEwV6bvHw/4ibkK5XirkmmSbQupjJTh
    jW2pOrnXM0xJ4EJ2WnkZLeV3bq0/Bj1OS8HV+fd9Xtk0XxM4takMymSLxO2bwRG1
    HvZly5yJtsfMwz4Sc0N2Sb+DzxsZYQKx1N2ya6ToKyaoyKCroK0jCPk=
    -----END RSA PRIVATE KEY-----"""
)

PUBLIC_KEY = inspect.cleandoc(
    """-----BEGIN PUBLIC KEY-----
    MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwA0cNMjzEm+LPFolbdOA
    cIB6hX7QGqkzVC8L8W5X2qFFDrSbTI9CSifQ3/1A+xYfiZrR4Za2cCvCtIzHxVct
    EAcQSW+IDRjh5vvdGTKgJMtaf3YHeXMk9NAN4Px6McuqaBE2llYt6okICha8MEUR
    9lffa761WkOOWnWAXQFUWxWOxuVnKQOOm1OQ0OiGd+6b81TbSmHDL1QeAMN66tsc
    nR0zFuLfaCRz8yZIsKR1bUPTAUnCB3gaUXIixxDTeSG6XHItb7c/uHJUPtW4lFLi
    DF12JX0GwQvkpYCe3I++fBxC0xcdGwmncKocOLtsuOPswDHH3waxp5KGVDsdx8Po
    8QIDAQAB
    -----END PUBLIC KEY-----"""
)

PRIVATE_SSH_KEY = inspect.cleandoc(
    """-----BEGIN OPENSSH PRIVATE KEY-----
    b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAACFwAAAAdzc2gtcn
    NhAAAAAwEAAQAAAgEA2Yilj6Eols7Ps51waZd73Kl+NQcYfrrFbCgeco4Y9RDwUA1bR/9A
    R18T+QzxfE3ZyjDWi+dcqd9yJDV9NvaKCen7HLJCrHOxCkaiHXTGP1l5Kt37gXWj1RPfip
    wLLx1nF13vccOR0phZmR7GhQc3XDocn3RwMN6sybsXVRUCqiFTJvKxn7ghHk9GsnZXEwg3
    xD+aKWWNHMRREoAgpxxKTxt08PO4zotGHJkEFWbtBcU4krs4RE0eQF36tzumMOXMvZ/efG
    eeKZMdJ+KQoD6bOXu5cHlwc93Vjo7KjheuIcGh8SkVXuXQ0Tn7bAtE2Qi5974R/jG5zd1c
    Seg1RJpRwWDqTgtl5uSdodWKkoBU57mFGkTVac3T+YieaeBp4KH1VFTqj/tuK1G93TfbcK
    iupRBDRv9FyGNV0OB/t3OMz7Pi3+4YCYUWhTm8PQx2cSTgSNARhEK0rK2P/yLhKccuqftM
    Aw2qFsVo7U4eIsnFKUAlYBJeRUeHsIPrc4U66Vn6q3zVjL4cvW5mwidG96ceXFCeNrM0Ai
    OyIApGAYoidImfisGL6xgwrT7yEJNK9521ewU5446J0lCfBXC4ewJhEPqFJSMN3e7eNNeU
    G6EJlwrptqTe9TstmPyXpnui/9zbQJzmE6aYD+cFcRWzGLfDeHi24dh39P5ttDabnLU/0g
    8AAAdQjgnT+I4J0/gAAAAHc3NoLXJzYQAAAgEA2Yilj6Eols7Ps51waZd73Kl+NQcYfrrF
    bCgeco4Y9RDwUA1bR/9AR18T+QzxfE3ZyjDWi+dcqd9yJDV9NvaKCen7HLJCrHOxCkaiHX
    TGP1l5Kt37gXWj1RPfipwLLx1nF13vccOR0phZmR7GhQc3XDocn3RwMN6sybsXVRUCqiFT
    JvKxn7ghHk9GsnZXEwg3xD+aKWWNHMRREoAgpxxKTxt08PO4zotGHJkEFWbtBcU4krs4RE
    0eQF36tzumMOXMvZ/efGeeKZMdJ+KQoD6bOXu5cHlwc93Vjo7KjheuIcGh8SkVXuXQ0Tn7
    bAtE2Qi5974R/jG5zd1cSeg1RJpRwWDqTgtl5uSdodWKkoBU57mFGkTVac3T+YieaeBp4K
    H1VFTqj/tuK1G93TfbcKiupRBDRv9FyGNV0OB/t3OMz7Pi3+4YCYUWhTm8PQx2cSTgSNAR
    hEK0rK2P/yLhKccuqftMAw2qFsVo7U4eIsnFKUAlYBJeRUeHsIPrc4U66Vn6q3zVjL4cvW
    5mwidG96ceXFCeNrM0AiOyIApGAYoidImfisGL6xgwrT7yEJNK9521ewU5446J0lCfBXC4
    ewJhEPqFJSMN3e7eNNeUG6EJlwrptqTe9TstmPyXpnui/9zbQJzmE6aYD+cFcRWzGLfDeH
    i24dh39P5ttDabnLU/0g8AAAADAQABAAACAAnI/8mKMk02WFl4B4p4afXvGeJAw3VP6XmI
    BHaatbywWl7/es+uR7HuMPPPBT3WMqnJsJzyd1Sc3IGMdml2hZcUk1sjGaSwP5J6wugYTK
    Kk37Jx8dJM/GWlUzTk+AvSgsSmURMI/RJAECe3Hq71OCRJ6OezUSihm5wnclE9dLU6cocZ
    +6t3mOXIIyH7+Ma+Y+PMCJ88FjDpE8yFPvLFbwJH9eawz+takHjQ2UItMxRDXtwa19karv
    WuRIzhfgRNEfylqFhoQNEV8QXORzsuFcVsL1dD3y8XF3gmLC5gGlEiaKp0cqWmKNKAjTK6
    /wnUPqlKP3zzt6vu/SvF6j9/wZWgAbOaRYn9r91YfsiiSBs+fnVnGJgFagMpgdy189StYl
    SOh9mIgjbUVpdvEE3oSgqLr9c38VoXfj4aoEGrxex/pRhFOtwe+Xy1K9ibYCALA2MRmA3O
    nOBGPD9BZDh0+smG63NuZ9Oxrh5iUhgnFMk5+E3D0yCM/xef35bQ6zV/gRpNly/X4IZk14
    vfeWXyS9wbc+ysldJIMLtT1oja01w4AmmPNT654qLwWFwCmXJi56dV4i9jlI4bG2MXoob5
    dKj9u+cOSfh6ENNWUvrIzvcM3pk5wHgTeTOBi6Ptb7iykk7/XEIXzxsyPyH4Sz8Oh/WMKs
    793bWJFs9xw29DBA2hAAABAQDzPGVsfLQj8n7OxHLag0GD+3ZvCoHcdgAFa17rpGg78I+x
    nglzYoNVJG3Wgia/gqg50W6uc2rhuvQ64u/N80mNuuAcrvdtS4ZO0fRUoLAiWN20xlphU+
    +CFhIjAqhbUo0QbNpPe/8K5JaxrappVJaQcN9oahi51je/qBmi2IShhb1bEIYQ9g3RZv+U
    QC8gnqReDO/bUFDPdogFe7Ks/rCeJR/rVCWsFZfxM4oc3wTSVqwagP6QFHwpuUVabEwkEc
    B3KCd8n25DxEKO6cVVcXQa9Ixe/lwZS7HWpTKbb7Zr4HBzX1LsyGIEon4QU5R6dKNs4g4A
    nMu1gmuyU9N+jndEAAABAQD9V+TzQJ2l0SXyoBDZNvzCN5XEsOZAerYhAGGN1UGcthwywm
    5KKQf7UJqS2vVzCcGIU0NePukwcNiq2PfST992uq34W99Hkhk8nOak597h/wCXsorfBbmN
    QeXw3a/Eaqn8P8pSc3dxuNIUUJdyoQ7qvRvuiYvPYXDTN2p5oGaNfRmmkWm79GqlBC1vT7
    QDbIP/SwtWUbIEio74I4rWB7qUD+vQBdYVkgiNw5CHp6QiCcr1iLChDrPF8RM83GunZblX
    +qmLPcoRWoAdAII6ECRRWzx5ZFC83m3R9cHVKyyq40r9q9tNr5Scp0PwPhtuBhSL84K1WP
    2IN4qXyaMbRIYTAAABAQDb0J7vwODCSOu0AxIPQkHtFOhnTp7RmdI1gmYf0d/ext2/ICMM
    /lXUZJilBk1OqHaoup3fgETmco0mKazewUG7G7KgXIZ6KiMpMDQVhIxuWUktVPPyj2/nSz
    CsU+b5zNBCsmZru9LzqSdYVBpZJMVXAtyeBas3kpHBcc3zddQw/1K7Y/gEjph8c4oJIuhP
    01Rh/sBVX3V2FCge0zTve7if4iepB9MWpd2/IwGy76iN5zz0HoFGwXxbT0ZY4sfNSwVzkg
    5VYQgNVhSsIkP4thLelIwzqx0HwCW3u3guqU+yOoPkamBJhTmF6RllnKTn3xOs0a+qLk1c
    0x7NdogmuDOVAAAAGXRlc3Rfb3Blbl9zc2hAZXhhbXBsZS5jb20B
    -----END OPENSSH PRIVATE KEY-----"""
)

PUBLIC_SSH_KEY = (
    "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQDZiKWPoSiWzs+znXBpl3vcq"
    "X41Bxh+usVsKB5yjhj1EPBQDVtH/0BHXxP5DPF8TdnKMNaL51yp33IkNX029o"
    "oJ6fscskKsc7EKRqIddMY/WXkq3fuBdaPVE9+KnAsvHWcXXe9xw5HSmFmZHsa"
    "FBzdcOhyfdHAw3qzJuxdVFQKqIVMm8rGfuCEeT0aydlcTCDfEP5opZY0cxFES"
    "gCCnHEpPG3Tw87jOi0YcmQQVZu0FxTiSuzhETR5AXfq3O6Yw5cy9n958Z54pk"
    "x0n4pCgPps5e7lweXBz3dWOjsqOF64hwaHxKRVe5dDROftsC0TZCLn3vhH+Mb"
    "nN3VxJ6DVEmlHBYOpOC2Xm5J2h1YqSgFTnuYUaRNVpzdP5iJ5p4GngofVUVOq"
    "P+24rUb3dN9twqK6lEENG/0XIY1XQ4H+3c4zPs+Lf7hgJhRaFObw9DHZxJOBI"
    "0BGEQrSsrY//IuEpxy6p+0wDDaoWxWjtTh4iycUpQCVgEl5FR4ewg+tzhTrpW"
    "fqrfNWMvhy9bmbCJ0b3px5cUJ42szQCI7IgCkYBiiJ0iZ+KwYvrGDCtPvIQk0"
    "r3nbV7BTnjjonSUJ8FcLh7AmEQ+oUlIw3d7t4015QboQmXCum2pN71Oy2Y/Je"
    "me6L/3NtAnOYTppgP5wVxFbMYt8N4eLbh2Hf0/m20NpuctT/SDw=="
    " test_open_ssh@example.com"
)
TABLE_NAME: str = "test_table"


def create_dataset(
    classification: bool = True,
    seed: int = 420,
    image: bool = False,
    multiimage: bool = False,
    multihead: bool = False,
    img_size: int = 50,
) -> pd.DataFrame:
    """Creates a random (seeded) dataset for testing."""
    seed_all(seed)

    base_date = datetime.date(2019, 1, 1)
    date_list = [
        base_date - datetime.timedelta(days=x) for x in range(int(DIMS[0] / 3))
    ] * 3
    df_date = pd.DataFrame(date_list, columns=["Date"])
    df_int = pd.DataFrame(np.random.randint(1, 1000, size=DIMS), columns=list("ABCD"))
    df_float = pd.DataFrame(np.random.uniform(0, 1, size=DIMS), columns=list("EFGH"))
    df_str = pd.DataFrame(
        np.random.choice(list(string.ascii_lowercase), size=DIMS).tolist(),
        columns=list("IJKL"),
    )
    df_bool = pd.DataFrame(
        list(np.random.choice([True, False], size=DIMS)), columns=list("MNOP")
    )
    target = pd.DataFrame(list(np.zeros(shape=DIMS[0], dtype=int)), columns=["TARGET"])

    data: pd.DataFrame = pd.concat(
        [df_date, df_int, df_float, df_str, df_bool, target], axis=1
    )

    if classification:
        data.loc[(data.A > 500) & (data.F < 0.5) & (data.D % 2 == 0), "TARGET"] = 1
    else:
        data["TARGET"] = data.E + data.F - (1 / data.B) / np.random.uniform()
    size = img_size if img_size is not None else 50
    if image or multiimage:

        def create_image(x: pd.DataFrame, size: int = size) -> BytesIO:
            def sigmoid(n: pd.Series) -> int:
                return int(255 / (1 + np.exp(-n)))

            # Use column "M" as defining whether the image is grayscale
            R, G, B = sigmoid(x["A"] / 100), sigmoid(x["F"]), sigmoid(x["D"] / 100)
            image = Image.new("RGB", size=(size, size), color=(R, G, B))
            if x["M"]:
                image = ImageOps.grayscale(image)
            temp = BytesIO()
            image.save(temp, format="png")
            return temp

        if image:
            data["image"] = data.apply(create_image, args=[size], axis=1)
        if multiimage:
            data["image1"] = data.apply(create_image, args=[size], axis=1)
            data["image2"] = data.apply(create_image, args=[size], axis=1)
    if multihead:
        data = data.assign(category=list(np.random.choice(["A", "B"], size=len(data))))
    return data


def create_dataset_pii(
    seed: int = 420,
) -> pd.DataFrame:
    """Creates a PII dataset with (seeded) random values for testing."""
    seed_all(seed)

    # Create a list of private identifiers
    df_names = pd.DataFrame(
        np.asarray(range(1, 5001)),
        columns=["name"],
    )
    df_int = pd.DataFrame(
        np.random.randint(1, 1000, size=(5000, 1)), columns=["int_field"]
    ).astype("int")
    df_str = pd.DataFrame(
        np.random.choice(list(string.ascii_lowercase), size=(5000, 1)).tolist(),
        columns=["str_field"],
    ).astype("str")
    df_obj = pd.DataFrame(
        np.random.choice(list(string.ascii_lowercase), size=(5000, 1)).tolist(),
        columns=["obj_field"],
    ).astype("object")

    df_age = pd.DataFrame(np.random.randint(1, 110, size=(5000, 1)), columns=["age"])
    df_height = pd.DataFrame(
        np.random.triangular(left=25, mode=168, right=230, size=(5000, 1)),
        columns=["height"],
    )
    df_weight = pd.DataFrame(
        np.random.triangular(left=1.8, mode=70, right=230, size=(5000, 1)),
        columns=["weight"],
    )

    exercise_choices = [
        "none",
        "run",
        "cycle",
        "swim",
        "row",
        "weights",
        "elliptical",
        "yoga",
        "pilates",
        "sport",
        "other",
    ]
    exercise_priors = [0.2, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08]
    df_exercise = pd.DataFrame(
        np.random.choice(exercise_choices, size=(5000, 1), p=exercise_priors),
        columns=["exercise"],
    )
    data: pd.DataFrame = pd.concat(
        [df_names, df_age, df_weight, df_height, df_exercise, df_int, df_str, df_obj],
        axis=1,
    )

    return data


def create_datastructure(
    multilabel: bool = False,
    multihead: bool = False,
    multihead_size: int = 1,
    loss_weights: bool = False,
) -> DataStructure:
    """This method creates and returns a DataStructure object."""
    target: Union[str, List[str]] = "TARGET"
    ignore_cols = ["Date"]
    loss_weights_col = None
    multihead_col = None

    if loss_weights:
        loss_weights_col = "weights"

    if multihead:
        multihead_col = "category"

    if multilabel:
        target = [cast(str, target), "TARGET_2"]

    return DataStructure(
        table=TABLE_NAME,
        target=target,
        ignore_cols=ignore_cols,
        multihead_col=multihead_col,
        multihead_size=multihead_size,
        loss_weights_col=loss_weights_col,
    )


def create_query_datastructure(
    multilabel: bool = False,
    multihead: bool = False,
    multihead_size: int = 1,
    loss_weights: bool = False,
) -> DataStructure:
    """This method creates and returns a DataStructure object."""
    target: Union[str, List[str]] = "target"
    loss_weights_col = None
    multihead_col = None
    query = """SELECT dd1."A" as a1, dd1."D" as d1, dd1."M" as m1,
                dd1."N" as n1, dd1."TARGET" as target,
                dd2."A" as a2, dd2."D" as d2, dd2."M" as m2, dd2."N" as n2
                FROM dummy_data dd1, dummy_data_2 dd2
                WHERE dd1."Date" = dd2."Date"
    """
    feature_override: SchemaOverrideMapping = {
        "continuous": ["a1", "d1", "a2", "d2"],
        "categorical": [
            {"m1": {"True": 1, "False": 0}},
            {"n1": {"True": 1, "False": 0}},
            {"m2": {"True": 1, "False": 0}},
            {"n2": {"True": 1, "False": 0}},
            {"target": {"0": 0, "1": 1}},
        ],
    }
    if loss_weights:
        loss_weights_col = "weights"

    if multihead:
        multihead_col = "category"

    if multilabel:
        target = [cast(str, target), "TARGET_2"]

    return DataStructure(
        query=query,
        schema_types_override=feature_override,
        target=target,
        multihead_col=multihead_col,
        multihead_size=multihead_size,
        loss_weights_col=loss_weights_col,
    )


def create_datasource(
    classification: bool,
    multilabel: bool = False,
    multihead: bool = False,
    loss_weights: bool = False,
    image: bool = False,
    multiimage: bool = False,
) -> BaseSource:
    """This method creates and returns a BaseSource object."""
    image_col: Optional[List[str]]
    if image:
        data = create_dataset(
            classification=classification, image=image, multihead=multihead
        )
        image_col = ["image"]
    elif multiimage:
        data = create_dataset(
            classification=classification, multiimage=multiimage, multihead=multihead
        )
        image_col = ["image1", "image2"]
    else:
        data = create_dataset(classification=classification, multihead=multihead)
        image_col = None

    if loss_weights:
        data = data.assign(weights=1.0)

    if multihead:
        data = data.assign(category=list(np.random.choice(["A", "B"], size=len(data))))

    if multilabel:
        data = data.assign(TARGET_2=np.zeros(len(data)))
        data.loc[(data.A < 700) & (data.F < 0.5) & (data.D % 2 == 1), "TARGET_2"] = 1

    dataset = DataFrameSource(data, seed=420, image_col=image_col)

    return dataset


def create_datasource_pii() -> BaseSource:
    """This method creates and returns a PII DataFrameSource dataset."""
    data = create_dataset_pii()

    dataset = DataFrameSource(data, seed=420, image_col=None)

    return dataset


def create_schema(
    classification: bool,
    multilabel: bool = False,
    multihead: bool = False,
    loss_weights: bool = False,
) -> BitfountSchema:
    """Helper function for creating datastructure and datasource."""
    datastructure = create_datastructure(multilabel, multihead, loss_weights)
    datasource = create_datasource(classification, multilabel, multihead, loss_weights)
    schema = BitfountSchema()
    force_stypes = {}
    force_categorical: List[Optional[str]] = []
    if classification:
        force_categorical = _add_this_to_list(datastructure.target, force_categorical)
        if multihead:
            force_categorical = _add_this_to_list(
                datastructure.multihead_col, force_categorical
            )

    if force_categorical:
        force_stypes = {TABLE_NAME: {"categorical": force_categorical}}
    datasource.load_data()
    schema.add_datasource_tables(
        datasource=datasource,
        table_name=TABLE_NAME,
        force_stypes=force_stypes,  # type: ignore[arg-type] # Reason: this will always have the needed type after the ifs. # noqa: B950
    )

    return schema


def create_schema_pii() -> BitfountSchema:
    """Helper function for creating datastructure and datasource for a PII dataset."""
    datasource = create_datasource_pii()
    schema = BitfountSchema()

    datasource.load_data()
    schema.add_datasource_tables(
        datasource=datasource,
        table_name=TABLE_NAME,
        force_stypes=None,
    )

    return schema


def get_datastructure_and_datasource(
    classification: bool,
    multilabel: bool = False,
    multihead: bool = False,
    multihead_size: int = 1,
    loss_weights: bool = False,
) -> Tuple[DataStructure, BaseSource]:
    """Helper function for creating datastructure and datasource."""
    datastructure = create_datastructure(
        multilabel, multihead, multihead_size, loss_weights
    )
    datasource = create_datasource(classification, multilabel, multihead, loss_weights)
    return datastructure, datasource


def create_segmentation_dataset(
    seg_dir: Path, height: int = 100, width: int = 100, count: int = 25
) -> pd.DataFrame:
    """Create a segmentation dataset."""
    seg = ExampleSegmentationData()
    input_images, target_masks = seg.generate_data(height, width, count=count)
    input_images_rgb = [x.astype(np.uint8) for x in input_images]
    target_masks_rgb = [seg.masks_to_colorimg(x.astype(np.uint8)) for x in target_masks]
    img_names_list = []
    masks_names_list = []
    # Save images
    for i in range(count):
        im2 = Image.fromarray((input_images_rgb[i]).astype(np.uint8))
        im2.save(f"{seg_dir}/img_{i}.png")
        img_names_list.append(f"img_{i}.png")
    # Save masks
    for i in range(count):
        im2 = Image.fromarray((target_masks_rgb[i]).astype(np.uint8))
        im2.save(f"{seg_dir}/masks_{i}.png")
        masks_names_list.append(f"masks_{i}.png")

    # Create dataframe with image and masks locations
    df = pd.DataFrame(
        {
            "img": [str(seg_dir) + "/" + img_name for img_name in img_names_list],
            "masks": [str(seg_dir) + "/" + mask_name for mask_name in masks_names_list],
        },
        columns=["img", "masks"],
    )
    return df


def create_local_modeller_and_workers(
    model_name: str,
    protocol_name: str,
    algorithm_name: str,
    secure_aggregation: bool = False,
    early_stopping: bool = False,
    auto_eval: bool = True,
) -> Tuple[_Modeller, List[_Worker]]:
    """Creates a modeller and workers to run in tests."""
    pod_ids = ["Alice/AlicePod", "Bob/BobPod"]
    pod_public_keys = {
        pod_id: create_autospec(RSAPublicKey, instance=True) for pod_id in pod_ids
    }
    results_only: bool = bool(protocol_name == "ResultsOnly")

    # Seed everything for consistency
    seed_all(43)

    # Create default hyperparams
    hyperparams: _StrAnyDict = {
        "epochs": 1,
        "batch_size": 32,
        "optimizer": Optimizer("RAdam", {"lr": 0.001}),
    }

    # Add to hyperparams based on model type
    if "RandomForest" in model_name or "LogisticRegression" in model_name:
        if not results_only:
            raise ValueError(
                "Random Forest and Logistic Regression can only be run with algorithm "
                "ResultsOnly"
            )
        hyperparams = {}

    # Create base data variables
    data = create_dataset(classification=True)
    schema = BitfountSchema()
    pod_datasets = []

    if results_only:
        # ResultsOnly only works with one pod
        pod_ids = ["Alice/AlicePod"]

    # Create datasets for each pod and combine them in the schema
    for _ in pod_ids:
        dataset = DataFrameSource(
            cast(pd.DataFrame, data.sample(n=1000, random_state=random.randint(1, 101)))
        )
        pod_datasets.append(dataset)

        # to make sure modeller schema corresponds to the worker schema
        schema.add_datasource_tables(
            dataset,
            table_name=TABLE_NAME,
            force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
            ignore_cols={TABLE_NAME: ["Date"]},
        )
    data_structure = DataStructure(
        table=TABLE_NAME, target="TARGET", ignore_cols=["Date"]
    )

    # Create local message service object
    mailbox_queues = LocalMessageServiceSharedQueues()
    modeller_name = "modeller-name"
    modeller_mailbox_id = "modeller_mailbox_id"
    # Just use the same key for each as we're running local tests
    aes_encryption_key = b"aes_encryption_key"

    # Handle early stopping
    early_stopping_ = None
    if early_stopping:
        early_stopping_ = FederatedEarlyStopping("validation_loss", 1, 0.01)

    # Create model
    model_class = _MODELS[model_name]
    model = model_class(datastructure=data_structure, schema=schema, **hyperparams)
    # Create algorithm
    algorithm = _ALGORITHMS[AlgorithmType[algorithm_name].name](
        model=model,
        epochs_between_parameter_updates=1,
    )

    # Create protocol
    protocol_kwargs: _StrAnyDict = {}
    if protocol_name == "FederatedAveraging":
        assert isinstance(model, DistributedModelProtocol)
        if secure_aggregation:
            sec_share = SecureShare()
            aggregator: _BaseAggregatorFactory = SecureAggregator(
                secure_share=sec_share
            )
        else:
            aggregator = Aggregator()

        protocol_kwargs = {
            "epochs_between_parameter_updates": 1,
            "aggregator": aggregator,
            "auto_eval": auto_eval,
        }
    protocol = _PROTOCOLS[ProtocolType[protocol_name].name](
        early_stopping=early_stopping_, algorithm=algorithm, **protocol_kwargs
    )

    # Create workers directly
    workers = []
    serialized_protocol = protocol.dump()
    for pod_identifier, dataset in zip(pod_ids, pod_datasets):
        message_service = LocalMessageService(
            username=pod_identifier.split("/")[0],
            shared_queues=mailbox_queues,
            modeller_mailbox_id=modeller_mailbox_id,
            pod_ids=pod_ids,
        )

        worker_mailbox: _WorkerMailbox
        if secure_aggregation:
            worker_mailbox = _InterPodWorkerMailbox(
                pod_public_keys=pod_public_keys,
                private_key=create_autospec(RSAPrivateKey, instance=True),
                pod_identifier=pod_identifier,
                modeller_mailbox_id=modeller_mailbox_id,
                modeller_name=modeller_name,
                aes_encryption_key=aes_encryption_key,
                message_service=message_service,
                pod_mailbox_ids=message_service.worker_mailbox_ids,
                task_id="this-is-a-task-id",
            )
        else:
            worker_mailbox = _WorkerMailbox(
                pod_identifier=pod_identifier,
                modeller_mailbox_id=modeller_mailbox_id,
                modeller_name=modeller_name,
                aes_encryption_key=aes_encryption_key,
                message_service=message_service,
                pod_mailbox_ids=message_service.worker_mailbox_ids,
                task_id="this-is-a-task-id",
            )

        worker = _Worker(
            datasource=dataset,
            mailbox=worker_mailbox,
            bitfounthub=create_autospec(BitfountHub, instance=True),
            authorisation=_LocalAuthorisation(
                _PodResponseMessage(modeller_name, pod_identifier),
                serialized_protocol,
            ),
            pod_identifier=pod_identifier,
            serialized_protocol=serialized_protocol,
        )
        workers.append(worker)

    # Create modeller
    modeller = _Modeller(
        protocol=protocol,
        message_service=LocalMessageService(
            username=modeller_name,
            shared_queues=mailbox_queues,
            modeller_mailbox_id=modeller_mailbox_id,
            pod_ids=pod_ids,
        ),
        bitfounthub=create_autospec(BitfountHub, instance=True),
        identity_verification_method=IdentityVerificationMethod.KEYS,
        private_key=create_autospec(RSAPrivateKey, instance=True),
    )

    return modeller, workers


def assert_results(
    model: _BaseModel,
    test_preds: Optional[np.ndarray] = None,
    test_target: Optional[Union[Sequence[float], Sequence[int]]] = None,
) -> None:
    """Assert results of a fitted model.

    This method takes a fitted model, evaluates it and performs some checks
    on the output.
    """
    if test_preds is None or test_target is None:
        preds, target = model.evaluate()
        # TODO: [BIT-1604] Remove these assert statements once they become superfluous.
        assert isinstance(preds, np.ndarray)
        assert isinstance(target, np.ndarray)
    else:
        preds = test_preds
        target = np.asarray(test_target)
    metrics = MetricCollection.create_from_model(model)
    results = metrics.compute(target, preds)
    assert results is not None
    assert isinstance(results, dict)
    if isinstance(model, ClassifierMixIn):
        auc = results["AUC"]
        assert auc > AUC_THRESHOLD
        if not model.multilabel and model.n_classes == 2:
            assert len(results) == len(BINARY_CLASSIFICATION_METRICS)
        else:
            assert len(results) == len(MULTICLASS_CLASSIFICATION_METRICS)
    else:
        assert len(results) == len(REGRESSION_METRICS)
        mae = results["MAE"]
        assert mae < MAE_THRESHOLD


class CallArgsLike(Protocol):
    """A protocol that defines a unittest.mock.call_args-like interface."""

    args: List[Any]
    kwargs: _StrAnyDict


def get_arg_from_args_or_kwargs(
    call_args: CallArgsLike, args_idx: int, kwarg_name: str
) -> Any:
    """Retrieves a mock function call argument regardless of how it was passed.

    Will extract a mock function argument regardless of if it was supplied by
    arg or kwarg.

    Args:
        call_args: The result of call_args from a mock call.
        args_idx: The potential position in *args the argument would be, zero-indexed.
        kwarg_name: The potential name in **kwargs of the argument.

    Returns:
        The argument, if found, or None if no argument was found.
    """
    # Check kwargs first as they must be provided after args
    try:
        return call_args.kwargs[kwarg_name]
    except KeyError:
        pass

    # Otherwise check args
    try:
        return call_args.args[args_idx]
    except IndexError:
        pass

    # Otherwise, couldn't be found
    raise ValueError(f'Could not find "{kwarg_name}" argument in call_args')


# caplog helper methods
_LOG_LEVEL_NAMES = Union[
    Literal["DEBUG"],
    Literal["INFO"],
    Literal["WARNING"],
    Literal["ERROR"],
    Literal["CRITICAL"],
]


def get_logs_at_level(
    caplog_fixture: LogCaptureFixture,
    levelname: _LOG_LEVEL_NAMES,
    full_details: bool = False,
    and_higher: bool = False,
) -> str:
    """Returns all logs at a given level as a single string.

    Args:
        caplog_fixture: The caplog fixture to extract from.
        levelname: The name of the log level to retrieve messages of.
        full_details: Whether to extract only the log record message or the full details
                      (timestamp, exception info, etc.). Uses the default formatter
                      to get the full log record output.
        and_higher: Whether to get logs only at target level or at any level greater
                    than or equal to the target level.

    Returns:
        All the log messages at the target level (or higher) as a newline-joined
        string.
    """
    if not and_higher:
        level_records = [
            record for record in caplog_fixture.records if record.levelname == levelname
        ]
    else:
        level_records = [
            record
            for record in caplog_fixture.records
            if record.levelno >= logging.getLevelName(levelname)
        ]

    # Need to interpolate/format the messages
    if full_details:
        formatter = logging.Formatter()
        level_messages = [formatter.format(r) for r in level_records]
    else:
        level_messages = [r.getMessage() for r in level_records]
    return "\n".join(level_messages)


def get_debug_logs(
    caplog_fixture: LogCaptureFixture,
    full_details: bool = False,
    and_higher: bool = False,
) -> str:
    """Returns all logs at debug level as a single string.

    All the log messages are combined as a newline-joined string.

    See `get_logs_at_level` for more information.
    """
    return get_logs_at_level(caplog_fixture, "DEBUG", full_details, and_higher)


def get_info_logs(
    caplog_fixture: LogCaptureFixture,
    full_details: bool = False,
    and_higher: bool = False,
) -> str:
    """Returns all logs at info level as a single string.

    All the log messages are combined as a newline-joined string.

    See `get_logs_at_level` for more information.
    """
    return get_logs_at_level(caplog_fixture, "INFO", full_details, and_higher)


def get_warning_logs(
    caplog_fixture: LogCaptureFixture,
    full_details: bool = False,
    and_higher: bool = False,
) -> str:
    """Returns all logs at warning level as a single string.

    All the log messages are combined as a newline-joined string.

    See `get_logs_at_level` for more information.
    """
    return get_logs_at_level(caplog_fixture, "WARNING", full_details, and_higher)


def get_error_logs(
    caplog_fixture: LogCaptureFixture,
    full_details: bool = False,
    and_higher: bool = False,
) -> str:
    """Returns all logs at error level as a single string.

    All the log messages are combined as a newline-joined string.

    See `get_logs_at_level` for more information.
    """
    return get_logs_at_level(caplog_fixture, "ERROR", full_details, and_higher)


def get_critical_logs(
    caplog_fixture: LogCaptureFixture,
    full_details: bool = False,
    and_higher: bool = False,
) -> str:
    """Returns all logs at critical level as a single string.

    All the log messages are combined as a newline-joined string.

    See `get_logs_at_level` for more information.
    """
    return get_logs_at_level(caplog_fixture, "CRITICAL", full_details, and_higher)



"""Custom exceptions for the Bitfount codebase."""
from typing import List

__all__: List[str] = [
    "BitfountEngineError",
    "BitfountError",
]


class BitfountError(Exception):
    """Base exception class that all others should inherit from."""

    pass


class BitfountEngineError(BitfountError):
    """Exception for any issues relating to the backend engine."""

    pass



"""Classes and functions for handling interaction with online storage services."""
import logging
import os
from pathlib import Path
from typing import Any, BinaryIO, Final, List, Optional, Tuple, Union

import httpx
from httpx import HTTPError as HTTPXError
from httpx import Response as HTTPXResponse
import msgpack
import requests
from requests import HTTPError, RequestException, Response

from bitfount import web_utils
from bitfount.types import _S3PresignedPOSTFields, _S3PresignedPOSTURL, _S3PresignedURL

__all__: List[str] = []

logger = logging.getLogger(__name__)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950


# Status codes that we consider as OK responses for upload
_OK_UPLOAD_STATUS_CODES: Tuple[int, ...] = (200, 201, 203, 204)
# Status codes that we consider as OK responses for download
_OK_DOWNLOAD_STATUS_CODES: Tuple[int, ...] = (200,)

_DEFAULT_FILE_NAME: Final[str] = "bitfount_file"


def _handle_request_exception(
    exc: Union[RequestException, HTTPXError],
    response: Optional[Union[Response, HTTPXResponse]],
    err_message: str,
) -> None:
    """Handles request exceptions in storage functions.

    Converts httpx to requests-style exceptions.
    """
    # Explicit None check needed due to truthyness of Response objects
    if response is not None:
        err_message += f": ({response.status_code})"
        if response.text:
            err_message += f" {response.text}"
    # If no response object, use exception info instead if present
    elif exc_str := str(exc):
        err_message += f": ({exc_str})"

    # If the exception is a HTTPXError subclass, we need to handle it specifically
    # to convert to a requests Exception.
    if isinstance(exc, httpx.HTTPStatusError):
        raise HTTPError(
            err_message, request=exc.request, response=exc.response
        ) from exc
    if isinstance(exc, httpx.ConnectError):
        raise requests.ConnectionError(err_message, request=exc.request) from exc
    if isinstance(exc, HTTPXError):
        # If it's a different HTTPXError subclass, handle it by a RequestException
        raise RequestException(err_message) from exc

    # Otherwise raise same type as what was passed in
    raise type(exc)(err_message) from exc


async def _async_upload_to_s3(
    upload_url: _S3PresignedPOSTURL,
    presigned_fields: _S3PresignedPOSTFields,
    to_upload: Union[bytes, BinaryIO],
    file_name: str = _DEFAULT_FILE_NAME,
) -> None:
    """Async Base s3 upload function.

    Handles the upload and response.

    Raises:
        requests.HTTPError: If the response code is not ok for any reason.
        requests.ConnectionError: If there is an error in the connection.
        requests.RequestException: For any other request-related issues.
    """
    try:
        response = await web_utils.async_post(
            url=upload_url,
            data=presigned_fields,
            files={"file": (file_name, to_upload)},
            timeout=None,
        )
        response.raise_for_status()
        if response.status_code not in _OK_UPLOAD_STATUS_CODES:
            raise HTTPError(f"Response not OK; not one of {_OK_UPLOAD_STATUS_CODES}")
    except (HTTPError, HTTPXError) as ex:
        try:
            # noinspection PyUnboundLocalVariable
            err_response: Optional[HTTPXResponse] = response
        except NameError:
            # response variable wasn't created yet
            err_response = None
        _handle_request_exception(ex, err_response, "Issue uploading object to S3")


async def _async_upload_data_to_s3(
    upload_url: _S3PresignedPOSTURL,
    presigned_fields: _S3PresignedPOSTFields,
    data: Any,
) -> None:
    """Asynchronously uploads an object to an s3 presigned POST url.

    Data will be packed via msgpack automatically before upload.

    Args:
        upload_url: The s3 presigned URL to upload to.
        presigned_fields: The additional fields needed to upload to the presigned
            URL. These are provided when the presigned URL is created.
        data: The data to upload.

    Raises:
        requests.HTTPError: If the response code is not ok for any reason.
        requests.ConnectionError: If there is an error in the connection.
        requests.RequestException: For any other request-related issues.
    """
    packed_data: bytes = msgpack.dumps(data)
    await _async_upload_to_s3(upload_url, presigned_fields, packed_data)


def _upload_to_s3(
    upload_url: _S3PresignedPOSTURL,
    presigned_fields: _S3PresignedPOSTFields,
    to_upload: Union[bytes, BinaryIO],
    file_name: str = _DEFAULT_FILE_NAME,
) -> None:
    """Base s3 upload function.

    Handles the upload and response.

    Raises:
        requests.HTTPError: If the response code is not 200 or 201.
        requests.RequestException: If the response is not OK for another reason.
    """
    try:
        response = web_utils.post(
            url=upload_url,
            data=presigned_fields,
            files={"file": (file_name, to_upload)},
        )

        response.raise_for_status()

        if response.status_code not in _OK_UPLOAD_STATUS_CODES:
            raise HTTPError(f"Response not OK; not one of {_OK_UPLOAD_STATUS_CODES}")
    except RequestException as ex:
        try:
            # noinspection PyUnboundLocalVariable
            err_response: Optional[Response] = response
        except NameError:
            # response variable wasn't created yet
            err_response = None
        _handle_request_exception(ex, err_response, "Issue uploading object to S3")


def _upload_data_to_s3(
    upload_url: _S3PresignedPOSTURL,
    presigned_fields: _S3PresignedPOSTFields,
    data: Any,
) -> None:
    """Uploads an object to an s3 presigned POST url.

    Data will be packed via msgpack automatically before upload.

    Args:
        upload_url: The s3 presigned URL to upload to.
        presigned_fields: The additional fields needed to upload to the presigned
            URL. These are provided when the presigned URL is created.
        data: The data to upload.

    Raises:
        requests.HTTPError: If the response code is not 200 or 201.
        requests.RequestException: If the response is not OK for another reason.
    """
    packed_data: bytes = msgpack.dumps(data)
    _upload_to_s3(upload_url, presigned_fields, packed_data)


def _upload_file_to_s3(
    upload_url: _S3PresignedPOSTURL,
    presigned_fields: _S3PresignedPOSTFields,
    file_path: Optional[Union[str, os.PathLike]] = None,
    file_contents: Optional[Union[str, bytes]] = None,
    file_encoding: str = "utf-8",
    file_name: Optional[str] = None,
) -> None:
    """Uploads a file to an s3 presigned POST url.

    The file is read and uploaded in binary mode as this is recommended for
    `requests.post()`.

    Args:
        upload_url: The s3 presigned URL to upload to.
        presigned_fields: The additional fields needed to upload to the presigned
            URL. These are provided when the presigned URL is created.
        file_path: The path to the file to be uploaded. Cannot be provided if
            file_contents is.
        file_contents: The contents of a file to be uploaded. Cannot be provided
            if file_path is.
        file_encoding: The encoding to use to convert string file contents to bytes.
        file_name: The name to upload the file with. Optional, default will be extracted
            automatically if using file_path, and use a default filename if using
            file_contents.

    Raises:
        requests.HTTPError: If the response code is not 200 or 201.
        requests.RequestException: If the response is not OK for another reason.
    """
    if bool(file_path) == bool(file_contents):
        raise ValueError(
            "One of file_path and file_contents must be provided, but not both."
        )

    # Handle if file_path provided
    if file_path:
        file_path = Path(file_path)

        # Use file's name if one not provided
        if not file_name:
            file_name = file_path.name

        with open(file_path, "rb") as f:
            _upload_to_s3(upload_url, presigned_fields, f, file_name)

    # Handle if file_contents provided
    if file_contents:
        # Use default filename if one not provided
        if not file_name:
            file_name = _DEFAULT_FILE_NAME

        # Convert to bytes as this is what's supported in `requests.post()`
        if isinstance(file_contents, str):
            file_contents = file_contents.encode(file_encoding)

        _upload_to_s3(upload_url, presigned_fields, file_contents, file_name)


async def _async_download_from_s3(download_url: _S3PresignedURL) -> bytes:
    """Async Base s3 download function.

    Handles the download and response parsing.

    Raises:
        httpx.HTTPError: If the response code is not ok for any reason.
    """
    try:
        response = await web_utils.async_get(url=download_url, timeout=None)
        response.raise_for_status()
        if response.status_code not in _OK_DOWNLOAD_STATUS_CODES:
            raise HTTPError(f"Response not OK; not one of {_OK_DOWNLOAD_STATUS_CODES}")
    except (HTTPError, HTTPXError) as ex:
        try:
            # noinspection PyUnboundLocalVariable
            err_response: Optional[HTTPXResponse] = response
        except NameError:
            # response variable wasn't created yet
            err_response = None
        _handle_request_exception(
            ex, err_response, "Issue whilst retrieving data from S3"
        )
    return response.content


async def _async_download_data_from_s3(download_url: _S3PresignedURL) -> Any:
    """Downloads data from the specified s3 URL.

    Will unpack the data using msgpack.

    Args:
        download_url: The s3 URL to download the data from.

    Raises:
        httpx.HTTPError: If the response code is not ok for any reason.
    """
    data: bytes = await _async_download_from_s3(download_url)
    return msgpack.loads(data)


def _download_from_s3(download_url: _S3PresignedURL) -> bytes:
    """Base s3 download function.

    Handles the download and response parsing.

    Raises:
        requests.HTTPError: If the response code is not 200.
        requests.RequestException: If the response is not OK.
    """
    try:
        response = web_utils.get(url=download_url)

        response.raise_for_status()

        if response.status_code not in _OK_DOWNLOAD_STATUS_CODES:
            raise HTTPError(f"Response not OK; not one of {_OK_DOWNLOAD_STATUS_CODES}")
    except RequestException as ex:
        try:
            # noinspection PyUnboundLocalVariable
            err_response: Optional[Response] = response
        except NameError:
            # response variable wasn't created yet
            err_response = None
        _handle_request_exception(
            ex, err_response, "Issue whilst retrieving data from S3"
        )
    return response.content


def _download_data_from_s3(download_url: _S3PresignedURL) -> Any:
    """Downloads data from the specified s3 URL.

    Will unpack the data using msgpack.

    Args:
        download_url: The s3 URL to download the data from.

    Raises:
        requests.HTTPError: If the response code is not 200.
        requests.RequestException: If the response is not OK.
    """
    data: bytes = _download_from_s3(download_url)
    return msgpack.loads(data)


def _download_file_from_s3(
    download_url: _S3PresignedURL, encoding: Optional[str] = None
) -> Union[str, bytes]:
    """Downloads a file from the specified s3 URL.

    Args:
        download_url: The s3 URL to download the data from.
        encoding: Optional. A string encoding to use to decode the data.

    Returns:
        The file contents, as a string if `encoding` provided otherwise as bytes.

    Raises:
        requests.HTTPError: If the response code is not 200.
        requests.RequestException: If the response is not OK.
    """
    data: bytes = _download_from_s3(download_url)
    if encoding:
        return data.decode(encoding)
    else:
        return data


def _get_packed_data_object_size(data: Any) -> int:
    """Get the size of a packed object as it will be uploaded to S3.

    Args:
        data: The object that will be uploaded to S3.

    Returns:
        The size of the packed object in bytes.
    """
    packed: bytes = msgpack.dumps(data)
    return len(packed)



"""Dataclasses to hold the configuration details for the runners."""
from __future__ import annotations

from dataclasses import dataclass
import logging
import os
from pathlib import Path
import re
import typing
from typing import Dict, List, Mapping, MutableMapping, Optional, Union

import desert
from marshmallow import ValidationError, fields, validate
from marshmallow.validate import OneOf
from marshmallow_union import Union as M_Union

from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    _get_environment,
)
from bitfount.data.types import (
    DataPathModifiers,
    SchemaOverrideMapping,
    _SemanticTypeValue,
)
from bitfount.data.utils import DatabaseConnection
from bitfount.federated.authorisation_checkers import (
    DEFAULT_IDENTITY_VERIFICATION_METHOD,
    IDENTITY_VERIFICATION_METHODS,
)
from bitfount.federated.privacy.differential import DPModellerConfig, DPPodConfig
from bitfount.federated.transport.config import (
    _DEV_MESSAGE_SERVICE_PORT,
    _DEV_MESSAGE_SERVICE_TLS,
    _DEV_MESSAGE_SERVICE_URL,
    _STAGING_MESSAGE_SERVICE_URL,
    MessageServiceConfig,
)
from bitfount.hub.authentication_flow import _DEFAULT_USERNAME
from bitfount.hub.types import (
    _DEV_AM_URL,
    _DEV_HUB_URL,
    _DEV_IDP_URL,
    _PRODUCTION_IDP_URL,
    _STAGING_AM_URL,
    _STAGING_HUB_URL,
    _STAGING_IDP_URL,
    PRODUCTION_AM_URL,
    PRODUCTION_HUB_URL,
    APIKeys,
)
from bitfount.models.base_models import (
    EarlyStopping,
    LoggerConfig,
    Optimizer,
    Scheduler,
)
from bitfount.types import _JSONDict

logger = logging.getLogger(__name__)

# POD_NAME_REGEX = re.compile(r"[a-z\d]+(-[a-z\d]+)*")
# USERNAME_REGEX = re.compile(r"[a-z\d](?:[a-z\d]|-(?=[a-z\d])){0,38}")
# TODO: [BIT-1493] revert to regex above to disallow underscores
POD_NAME_REGEX = re.compile(r"[a-z\d]+((-|_)[a-z\d]+)*")
USERNAME_REGEX = re.compile(r"[a-z\d](?:[a-z\d]|(-|_)(?=[a-z\d])){0,38}")

# MYPY ISSUES
# Due to the use of desert.field to create non-standard fields in these dataclasses,
# mypy complains about incompatible type assignment. This is a known bug in desert
# (https://github.com/python-desert/desert/issues/101) and so we can ignore it
# until the issue is resolved.


def _deserialize_path(path: Optional[str]) -> Optional[Path]:
    """Converts a str into a Path.

    If the input is None, the output is None.
    """
    if path is None:
        return path
    else:
        return Path(path).expanduser()


def _deserialize_model_ref(ref: str) -> Union[Path, str]:
    """Deserializes a model reference.

    If the reference is a path to a file (and that file exists), return a Path
    instance. Otherwise, returns the str reference unchanged.
    """
    path = Path(ref).expanduser()
    if path.is_file():  # also returns False if path doesn't exist
        return path
    else:
        return ref


# COMMON SCHEMAS
@dataclass
class AccessManagerConfig:
    """Configuration for the access manager."""

    url: str = desert.field(  # type: ignore[assignment] # Reason: see top of file
        fields.URL(), default=PRODUCTION_AM_URL
    )


@dataclass
class HubConfig:
    """Configuration for the hub."""

    url: str = desert.field(  # type: ignore[assignment] # Reason: see top of file
        fields.URL(), default=PRODUCTION_HUB_URL
    )


# MODELLER SCHEMAS
@dataclass
class ModellerUserConfig:
    """Configuration for the modeller.

    Args:
        username: The username of the modeller. This can be picked up automatically
            from the session but can be overridden here.
        identity_verification_method: The method to use for identity verification.
            Accepts one of the values in `IDENTITY_VERIFICATION_METHODS`, i.e. one of
            `key-based`, `saml`, `oidc-auth-code` or `oidc-device-code`.
        private_key_file: The path to the private key file for key-based identity
            verification.
    """

    username: str = desert.field(
        fields.String(validate=validate.Regexp(USERNAME_REGEX)),
        default=_DEFAULT_USERNAME,
    )  # type: ignore[assignment] # Reason: see top of file

    identity_verification_method: str = desert.field(
        fields.String(validate=OneOf(IDENTITY_VERIFICATION_METHODS)),
        default=DEFAULT_IDENTITY_VERIFICATION_METHOD,
    )  # type: ignore[assignment] # Reason: see top of file
    private_key_file: Optional[Path] = desert.field(  # type: ignore[assignment] # Reason: see top of file  # noqa: B950
        fields.Function(deserialize=_deserialize_path), default=None
    )

    def __post_init__(self) -> None:
        environment = _get_environment()
        self._identity_provider_url: str

        if environment == _STAGING_ENVIRONMENT:
            self._identity_provider_url = _STAGING_IDP_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            self._identity_provider_url = _DEV_IDP_URL
        else:
            self._identity_provider_url = _PRODUCTION_IDP_URL


@dataclass
class ModellerConfig:
    """Full configuration for the modeller."""

    pods: PodsConfig
    data_structure: DataStructureConfig
    task: TaskConfig

    modeller: ModellerUserConfig = desert.field(  # type: ignore[assignment] # Reason: see top of file  # noqa: B950
        fields.Nested(desert.schema_class(ModellerUserConfig)),
        default_factory=ModellerUserConfig,
    )
    hub: HubConfig = desert.field(  # type: ignore[assignment] # Reason: see top of file
        fields.Nested(desert.schema_class(HubConfig)), default_factory=HubConfig
    )
    message_service: MessageServiceConfig = desert.field(
        fields.Nested(desert.schema_class(MessageServiceConfig)),
        default_factory=MessageServiceConfig,
    )  # type: ignore[assignment] # Reason: see top of file

    def __post_init__(self) -> None:
        environment = _get_environment()
        if environment == _STAGING_ENVIRONMENT:
            self.hub.url = _STAGING_HUB_URL
            self.message_service.url = _STAGING_MESSAGE_SERVICE_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            self.hub.url = _DEV_HUB_URL
            self.message_service.url = _DEV_MESSAGE_SERVICE_URL
            self.message_service.port = _DEV_MESSAGE_SERVICE_PORT
            self.message_service.tls = _DEV_MESSAGE_SERVICE_TLS


@dataclass
class ModelStructureConfig:
    """Configuration for the ModelStructure."""

    name: str
    arguments: _JSONDict = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file


@dataclass
class BitfountModelReferenceConfig:
    """Configuration for BitfountModelReference."""

    model_ref: Union[Path, str] = desert.field(
        fields.Function(deserialize=_deserialize_model_ref)
    )  # type: ignore[assignment] # Reason: see top of file
    model_version: Optional[int] = None
    username: Optional[str] = None


@dataclass
class ModelConfig:
    """Configuration for the model."""

    # For existing models
    name: Optional[str] = None
    structure: Optional[ModelStructureConfig] = None

    # For custom models
    bitfount_model: Optional[BitfountModelReferenceConfig] = None

    # Other
    hyperparameters: _JSONDict = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file
    logger_config: Optional[LoggerConfig] = None
    dp_config: Optional[DPModellerConfig] = None

    def __post_init__(self) -> None:
        # Validate either name or bitfount_model reference provided
        self._name_or_bitfount_model()
        try:
            self.hyperparameters["optimizer"] = Optimizer(
                **self.hyperparameters["optimizer"]
            )
        except KeyError:
            pass
        try:
            self.hyperparameters["early_stopping"] = EarlyStopping(
                **self.hyperparameters["early_stopping"]
            )
        except KeyError:
            pass
        try:
            self.hyperparameters["scheduler"] = Scheduler(
                **self.hyperparameters["scheduler"]
            )
        except KeyError:
            pass

    def _name_or_bitfount_model(self) -> None:
        """Ensures that both `name` and `bitfount_model` can't be set.

        Raises:
            ValidationError: if both `name` and `bitfount_model` are set
        """
        if self.name and self.bitfount_model:
            raise ValidationError("Cannot specify both name and bitfount_model.")
        if not self.name and not self.bitfount_model:
            raise ValidationError("Must specify either name or bitfount_model.")


@dataclass
class PodsConfig:
    """Configuration for the pods to use for the modeller."""

    identifiers: List[str]


@dataclass
class ProtocolConfig:
    """Configuration for the Protocol."""

    name: str
    arguments: _JSONDict = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file

    def __post_init__(self) -> None:
        # TODO: [BIT-1300] Remove this once custom protocols are supported
        # This is added for convenience so the user doesn't have to be verbose with the
        # protocol name when there are only bitfount protocols available
        if not self.name.startswith("bitfount."):
            self.name = f"bitfount.{self.name}"


@dataclass
class AggregatorConfig:
    """Configuration for the Aggregator."""

    secure: bool
    weights: Optional[Dict[str, Union[int, float]]] = None

    def __post_init__(self) -> None:
        if self.secure and self.weights:
            # TODO: [BIT-1486] Remove this constraint
            raise NotImplementedError(
                "SecureAggregation does not support update weighting"
            )


@dataclass
class AlgorithmConfig:
    """Configuration for the Algorithm."""

    name: str
    arguments: _JSONDict = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file
    pretrained_file: Optional[Path] = desert.field(
        fields.Function(deserialize=_deserialize_path), default=None
    )  # type: ignore[assignment] # Reason: see top of file

    def __post_init__(self) -> None:
        # TODO: [BIT-1300] Remove this once custom algorithms are supported
        # This is added for convenience so the user doesn't have to be verbose with the
        # algorithm name when there are only bitfount algorithms available
        if not self.name.startswith("bitfount."):
            self.name = f"bitfount.{self.name}"


@dataclass
class TaskConfig:
    """Configuration for the task."""

    protocol: ProtocolConfig
    algorithm: AlgorithmConfig
    model: ModelConfig
    aggregator: Optional[AggregatorConfig] = None
    transformation_file: Optional[Path] = desert.field(
        fields.Function(deserialize=_deserialize_path), default=None
    )  # type: ignore[assignment] # Reason: see top of file


# POD SCHEMAS
@dataclass
class PodConfig:
    """Full configuration for the pod.

    Raises:
        ValueError: If a username is not provided alongside API keys.
    """

    datasource: str
    data_config: PodDataConfig
    api_key: Optional[APIKeys] = None
    name: str = fields.String(
        validate=validate.Regexp(POD_NAME_REGEX)
    )  # type: ignore[assignment] # Reason: see top of file
    pod_details_config: Optional[PodDetailsConfig] = None
    schema: Optional[Path] = desert.field(
        fields.Function(deserialize=_deserialize_path), default=None
    )  # type: ignore[assignment] # Reason: see top of file
    access_manager: AccessManagerConfig = desert.field(
        fields.Nested(desert.schema_class(AccessManagerConfig)),
        default_factory=AccessManagerConfig,
    )  # type: ignore[assignment] # Reason: see top of file
    hub: HubConfig = desert.field(  # type: ignore[assignment] # Reason: see top of file
        fields.Nested(desert.schema_class(HubConfig)), default_factory=HubConfig
    )
    message_service: MessageServiceConfig = desert.field(
        fields.Nested(desert.schema_class(MessageServiceConfig)),
        default_factory=MessageServiceConfig,
    )  # type: ignore[assignment] # Reason: see top of file
    differential_privacy: Optional[DPPodConfig] = None
    approved_pods: Optional[List[str]] = None
    username: str = desert.field(
        fields.String(validate=validate.Regexp(USERNAME_REGEX)),
        default=_DEFAULT_USERNAME,
    )  # type: ignore[assignment] # Reason: see top of file

    def __post_init__(self) -> None:
        environment = _get_environment()
        if environment == _STAGING_ENVIRONMENT:
            self.hub.url = _STAGING_HUB_URL
            self.access_manager.url = _STAGING_AM_URL
            self.message_service.url = _STAGING_MESSAGE_SERVICE_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            self.hub.url = _DEV_HUB_URL
            self.access_manager.url = _DEV_AM_URL
            self.message_service.url = _DEV_MESSAGE_SERVICE_URL
            self.message_service.port = _DEV_MESSAGE_SERVICE_PORT
            self.message_service.tls = _DEV_MESSAGE_SERVICE_TLS

        # Use API Keys for authentication if provided
        if self.api_key:
            if self.username == _DEFAULT_USERNAME:
                raise ValueError("Must specify a username when using API Keys.")

            logger.info("Setting API Keys as environment variables.")

            if os.environ.get("BITFOUNT_API_KEY_ID") or os.environ.get(
                "BITFOUNT_API_KEY"
            ):
                logger.warning(
                    "Existing environment variable API keys detected. Overriding with "
                    "those provided in the pod config."
                )
            os.environ["BITFOUNT_API_KEY_ID"] = self.api_key.access_key_id
            os.environ["BITFOUNT_API_KEY"] = self.api_key.access_key

    @property
    def pod_id(self) -> str:
        """The pod ID of the pod specified."""
        return f"{self.username}/{self.name}"


@dataclass
class PathConfig:
    """Configuration for the path."""

    path: Path = desert.field(
        fields.Function(deserialize=_deserialize_path)
    )  # type: ignore[assignment] # Reason: see top of file


@dataclass
class DatabaseConfig:
    """Configuration for the DatabaseConnection."""

    # This dataclass is a mostly superfluous wrapper around DatabaseConnection.
    # It is needed due to a bug in desert (https://github.com/python-desert/desert/issues/122)  # noqa: B950
    # which causes explicitly Optional Union fields to error out. Our mypy config
    # requires explicit Optional for good practice so rather than corrupt the
    # existing dataclass it makes sense to include a mirror config here so this
    # entire module can be excluded from that mypy check.
    #
    # Once the bug is fixed we can remove this class and replace it with
    # DatabaseConnection directly.

    con: str
    db_schema: Optional[str] = None
    query: Optional[str] = None
    table_names: Optional[List[str]] = None

    def get_db_connection(self) -> DatabaseConnection:
        """Returns a DatabaseConnection."""
        return DatabaseConnection(
            con=self.con,
            db_schema=self.db_schema,
            query=self.query,
            table_names=self.table_names,
        )


@dataclass
class DataSplitConfig:
    """Configuration for the data splitter."""

    data_splitter: str = desert.field(
        fields.String(validate=OneOf(["percentage", "predefined"])),
        default="percentage",
    )  # type: ignore[assignment] # Reason: see top of file
    args: _JSONDict = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file


@dataclass
class PodDataConfig:
    """Configuration for the Schema, BaseSource and Pod.

    Args:
        force_stypes: The semantic types to force for the data. This is passed to the
            `BitfountSchema`. This should be a mapping from pod name to a mapping from
            the type to a list of column names (e.g. {"pod_name": {categorical:
            ["col1", "col2"]}}, or a mappping from each table name in a multitable
            datasource to a mapping from the type to a list of column names (e.g.
            {"table1": {categorical: ["col1", "col2"]}, "table2":  {continuous:
            ["col3", "col4"]}}.
        ignore_cols: The columns to ignore. This is passed to the data source.
        modifiers: The modifiers to apply to the data. This is passed to the
            `BaseSource`.
        datasource_args: Key-value pairs of arguments to pass to the data source
            constructor.
        data_split: The data split configuration. This is passed to the data source.
        auto_tidy: Whether to automatically tidy the data. This is used by the
            `Pod` and will result in removal of NaNs and normalisation of numeric
            values. Defaults to False.
    """

    # There is a bug in desert (https://github.com/python-desert/desert/issues/122)
    # which causes explicitly Optional Union fields to error out. Our mypy config
    # requires explicit Optional for good practice so we currently have to do more
    # involved desert.field() assignments until the issue is fixed.

    force_stypes: Optional[
        Mapping[str, MutableMapping[_SemanticTypeValue, List[str]]]
    ] = desert.field(
        fields.Dict(
            keys=fields.String(),
            values=fields.Dict(
                keys=fields.String(validate=OneOf(typing.get_args(_SemanticTypeValue))),
                values=fields.List(fields.String()),
            ),
        ),
        default=None,
    )  # type: ignore[assignment] # Reason: see comment and top of file

    ignore_cols: Optional[Mapping[str, List[str]]] = desert.field(
        fields.Dict(keys=fields.String(), values=fields.List(fields.String())),
        default=None,
    )  # type: ignore[assignment] # Reason: see comment and top of file

    modifiers: Optional[Dict[str, DataPathModifiers]] = desert.field(
        fields.Dict(
            keys=fields.Str,
            values=fields.Dict(
                keys=fields.String(
                    validate=OneOf(DataPathModifiers.__annotations__.keys())
                )
            ),
            default=None,
        ),
        default=None,
    )  # type: ignore[assignment] # Reason: see comment and top of file
    datasource_args: _JSONDict = desert.field(
        fields.Dict(keys=fields.Str), default_factory=dict
    )  # type: ignore[assignment] # Reason: see top of file
    data_split: DataSplitConfig = desert.field(
        fields.Nested(desert.schema_class(DataSplitConfig)),
        default_factory=DataSplitConfig,
    )  # type: ignore[assignment] # Reason: see top of file
    auto_tidy: bool = False


@dataclass
class PodDetailsConfig:
    """Configuration for the pod details.

    Args:
        display_name: The display name of the pod.
        description: The description of the pod.
    """

    display_name: str
    description: str


@dataclass
class DataStructureSelectConfig:
    """Configuration for the datastructure select argument."""

    include: List[str] = desert.field(
        fields.List(fields.String()), default_factory=list
    )  # type: ignore[assignment] # Reason: see top of file
    exclude: List[str] = desert.field(
        fields.List(fields.String()), default_factory=list
    )  # type: ignore[assignment] # Reason: see top of file


@dataclass
class DataStructureAssignConfig:
    """Configuration for the datastructure assign argument."""

    target: Union[str, List[str]] = desert.field(
        M_Union([fields.String(), fields.List(fields.String())])
    )  # type: ignore[assignment] # Reason: see top of file
    # When we add more DataStructures, we should make the target an optional argument
    image_cols: Optional[List[str]] = None
    loss_weights_col: Optional[str] = None
    multihead_col: Optional[str] = None
    ignore_classes_col: Optional[str] = None


@dataclass
class DataStructureTransformConfig:
    """Configuration for the datastructure transform argument."""

    dataset: Optional[List[Dict[str, _JSONDict]]] = None
    batch: Optional[List[Dict[str, _JSONDict]]] = None


@dataclass
class DataStructureTableConfig:
    """Configuration for the datastructure table arguments."""

    table: Optional[Union[str, Dict[str, str]]] = desert.field(
        M_Union(
            [
                fields.String(),
                fields.Dict(keys=fields.String(), values=fields.String(), default=None),
            ],
            default=None,
        ),
        default=None,
    )  # type: ignore[assignment] # Reason: see top of file
    query: Optional[Union[str, Dict[str, str]]] = desert.field(
        M_Union(
            [
                fields.String(),
                fields.Dict(keys=fields.String(), values=fields.String(), default=None),
            ],
            default=None,
        ),
        default=None,
    )  # type: ignore[assignment] # Reason: see top of file
    schema_types_override: Optional[
        Union[SchemaOverrideMapping, Mapping[str, SchemaOverrideMapping]]
    ] = desert.field(
        fields.Dict(
            keys=fields.String(),
            values=fields.List(
                M_Union([fields.String(default=None), fields.Dict(default=None)])
            ),
            default=None,
        ),
        default=None,
    )  # type: ignore[assignment] # Reason: see top of file


@dataclass
class DataStructureConfig:
    """Configuration for the modeller schema and dataset options."""

    table_config: DataStructureTableConfig = desert.field(
        fields.Nested(desert.schema_class(DataStructureTableConfig)),
    )  # type: ignore[assignment] # Reason: see top of file
    assign: DataStructureAssignConfig = desert.field(
        fields.Nested(desert.schema_class(DataStructureAssignConfig)),
    )  # type: ignore[assignment] # Reason: see top of file
    select: DataStructureSelectConfig = desert.field(
        fields.Nested(desert.schema_class(DataStructureSelectConfig)),
        default_factory=DataStructureSelectConfig,
    )  # type: ignore[assignment] # Reason: see top of file
    transform: DataStructureTransformConfig = desert.field(
        fields.Nested(desert.schema_class(DataStructureTransformConfig)),
        default_factory=DataStructureTransformConfig,
    )  # type: ignore[assignment] # Reason: see top of file



"""Distributed tests with pytorch models."""

from typing import Optional

import pytest
import pytorch_lightning as pl
import torch

from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.backends.pytorch.types import _AdaptorForPyTorchTensor
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from tests.bitfount.backends.pytorch.helper import get_params_mean
from tests.utils.helper import (
    backend_test,
    create_datasource,
    create_datastructure,
    unit_test,
)


@pytest.fixture
def datastructure() -> DataStructure:
    """Fixture for datastructure."""
    return create_datastructure()


@pytest.fixture
def datasource(datastructure: DataStructure) -> BaseSource:
    """Fixture for datasource."""
    return create_datasource(classification=True)


@backend_test
class TestDistributedModel:
    """Test distributed model methods with PyTorch models."""

    @unit_test
    def test_set_model_training_iterations_epochs(
        self, datastructure: DataStructure
    ) -> None:
        """Test setting of model training iterations with epochs."""
        model = PyTorchTabularClassifier(
            datastructure=datastructure, schema=BitfountSchema(), seed=42, epochs=2
        )
        model.set_model_training_iterations(7)
        assert model.epochs == 7
        assert model.steps is None

    @unit_test
    def test_set_model_training_iterations_steps(
        self, datastructure: DataStructure
    ) -> None:
        """Test setting of model training iterations with steps."""
        model = PyTorchTabularClassifier(
            datastructure=datastructure, schema=BitfountSchema(), seed=42, steps=122
        )
        model.set_model_training_iterations(7)
        assert model.steps == 7
        assert model.epochs is None

    @unit_test
    def test_model_update_params(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Check that updating model params works properly."""
        model1 = PyTorchTabularClassifier(
            datastructure=datastructure, schema=BitfountSchema(), seed=42, epochs=2
        )
        model1.initialise_model(datasource)
        model1_original_params = model1.get_param_states()

        model2 = PyTorchTabularClassifier(
            datastructure=datastructure, schema=BitfountSchema(), seed=43, epochs=2
        )
        model2.initialise_model(datasource)
        model2_original_params = model2.get_param_states()

        assert not torch.isclose(
            get_params_mean(model1_original_params),
            get_params_mean(model2_original_params),
            atol=1e-4,
        )
        model1.update_params(model2_original_params)
        assert torch.isclose(
            get_params_mean(model1.get_param_states()),
            get_params_mean(model2_original_params),
            atol=1e-4,
        )

    @unit_test
    def test_model_diff_params(self) -> None:
        """Check that diffing model params works properly."""
        old_params = {
            "layer1": _AdaptorForPyTorchTensor(torch.tensor([1.0, 2.0, 3.0, 4.0]))
        }
        new_params = {
            "layer1": _AdaptorForPyTorchTensor(torch.tensor([10.0, 20.0, 30.0, 40.0]))
        }
        expected_param_diff = torch.tensor([9.0, 18.0, 27.0, 36.0])
        param_diff = PyTorchTabularClassifier.diff_params(old_params, new_params)
        if isinstance(param_diff["layer1"], _AdaptorForPyTorchTensor):
            for i, j in zip(param_diff["layer1"].torchtensor, expected_param_diff):
                assert torch.isclose(i, j)

    @unit_test
    @pytest.mark.parametrize("epochs,steps", [(1, None), (None, 1)])
    def test_reset_trainer(self, epochs: Optional[int], steps: Optional[int]) -> None:
        """Tests resetting of trainer."""
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            epochs=epochs,
            steps=steps,
        )
        model.initialise_model(create_datasource(classification=True))
        model._pl_trainer.random_attribute = 5  # type: ignore[attr-defined] # Reason: integral to test # noqa: B950
        model._total_num_batches_trained = 5
        model.reset_trainer()
        assert isinstance(model._pl_trainer, pl.Trainer)
        # Check that new trainer does not have the random attribute to ensure it is new
        assert not hasattr(model._pl_trainer, "random_attribute")
        # Check that the number of steps to be trained is updated correctly
        if steps:
            assert model.steps == steps
            assert (
                model._pl_trainer.max_steps == model._total_num_batches_trained + steps
            )
        # Check that the number of epochs to be trained is updated correctly
        elif epochs:
            assert model.epochs == epochs
            assert model._pl_trainer.max_epochs == epochs

    @unit_test
    @pytest.mark.parametrize(
        "steps,total_num_batches_trained", [(1, 0), (3, 3), (5, 10)]
    )
    def test_reset_trainer_steps(
        self, steps: int, total_num_batches_trained: int
    ) -> None:
        """Tests resetting of trainer with steps specfied as the type of iteration.

        Checks that the `max_steps` attribute on the trainer is correctly set.
        """
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            steps=steps,
        )
        model.initialise_model(create_datasource(classification=True))
        model._total_num_batches_trained = total_num_batches_trained
        model.reset_trainer()

        assert model._pl_trainer.max_steps == steps + (
            total_num_batches_trained % len(model.train_dl)
        )



"""Tests for the sql query algorithm."""
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, create_autospec

from _pytest.logging import LogCaptureFixture
from _pytest.tmpdir import TempPathFactory
import numpy as np
import pandas as pd
from pytest import fixture
from pytest_mock import MockerFixture
import sqlalchemy

from bitfount import DatabaseConnection
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datasources.excel_source import ExcelSource
from bitfount.data.datasources.intermine_source import IntermineSource
from bitfount.federated.algorithms.base import (
    _BaseAlgorithm,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.algorithms.sql_query import SqlQuery, _ModellerSide, _WorkerSide
from bitfount.federated.modeller import _Modeller
from bitfount.federated.utils import _ALGORITHMS
from bitfount.hub import BitfountHub
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import (
    create_dataset,
    create_datasource,
    integration_test,
    unit_test,
)


class TestSqlQuery:
    """Test SqlQuery algorithm."""

    @fixture
    def datasource(self) -> BaseSource:
        """Fixture for datasource."""
        return create_datasource(classification=True)

    @fixture(scope="class")
    def multi_table_excel_source(self, multi_table_excel_file: Path) -> ExcelSource:
        """Multi Table ExcelSource."""
        source = ExcelSource(multi_table_excel_file, sheet_name=["Sheet1", "Sheet2"])
        assert source.multi_table
        return source

    @fixture(scope="class")
    def multi_table_excel_file(self, tmp_path_factory: TempPathFactory) -> Path:
        """Path to multi table excel file."""
        dataframe = create_dataset()
        tmp_path = tmp_path_factory.mktemp("temp_excel")
        filename = tmp_path / "test.xlsx"
        with pd.ExcelWriter(filename) as writer:  # type: ignore[abstract] # Reason: This is the documented usage. # noqa: B950
            dataframe.to_excel(writer, index=False, sheet_name="Sheet1")
            dataframe.to_excel(writer, index=False, sheet_name="Sheet2")

        return filename

    @unit_test
    def test_modeller_types(self) -> None:
        """Test modeller method."""
        algorithm_factory = SqlQuery(query="SELECT * from `df`")
        algorithm = algorithm_factory.modeller()
        for type_ in [
            _BaseAlgorithm,
            _BaseModellerAlgorithm,
        ]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker_types(self) -> None:
        """Test worker method."""
        algorithm_factory = SqlQuery(query="SELECT * from `df`")
        algorithm = algorithm_factory.worker(
            hub=create_autospec(BitfountHub, instance=True)
        )
        for type_ in [
            _BaseAlgorithm,
            _BaseWorkerAlgorithm,
        ]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker_init_datasource(self, datasource: DataFrameSource) -> None:
        """Test worker method."""
        algorithm_factory = SqlQuery(query="SELECT * from `df`")
        algorithm_factory.worker().initialise(datasource=datasource)

    @unit_test
    def test_modeller_init(self) -> None:
        """Test worker method."""
        algorithm_factory = SqlQuery(query="SELECT * from `df`")
        algorithm_factory.modeller().initialise()

    @unit_test
    def test_bad_sql_no_table(
        self, caplog: LogCaptureFixture, datasource: DataFrameSource
    ) -> None:
        """Test that having no data raises an error."""
        algorithm_factory = SqlQuery(query="SELECT MAX(G) AS MAX_OF_G", table="df")
        worker = algorithm_factory.worker()
        worker.datasource = datasource
        worker.datasource.load_data()
        worker.run()
        err_msg = "The default table for single table datasource is the pod"
        assert err_msg in caplog.text

    @unit_test
    def test_bad_sql_query_statement(
        self, caplog: LogCaptureFixture, datasource: DataFrameSource
    ) -> None:
        """Test that a bad operator in SQL query errors out."""
        algorithm_factory = SqlQuery(query="SELECTOR MAX(G) AS MAX_OF_G FROM df")
        worker = algorithm_factory.worker()
        worker.datasource = datasource
        worker.datasource.load_data()
        worker.run()
        assert "ERROR" in caplog.text

    @unit_test
    def test_bad_sql_query_column(
        self, caplog: LogCaptureFixture, datasource: DataFrameSource
    ) -> None:
        """Test that an invalid column in SQL query errors out."""
        algorithm_factory = SqlQuery(
            query="SELECT MAX(BITFOUNT_TEST) AS MAX_OF_BITFOUNT_TEST FROM `df`",
            table="df",
        )
        worker = algorithm_factory.worker()
        worker.datasource = datasource
        worker.run()
        assert "ERROR" in caplog.text

    @unit_test
    def test_bad_sql_no_from_df(
        self, caplog: LogCaptureFixture, datasource: DataFrameSource
    ) -> None:
        """Test that an invalid query errors out."""
        algorithm_factory = SqlQuery(query="mock", table="df")
        worker = algorithm_factory.worker()
        worker.datasource = datasource
        worker.run()
        err_msg = "The default table for single table datasource is the pod"
        assert err_msg in caplog.text

    @unit_test
    def test_worker_gets_sql_results(self, datasource: DataFrameSource) -> None:
        """Test that a SQL query returns correct result."""
        algorithm_factory = SqlQuery(
            query="SELECT MAX(G) AS MAX_OF_G FROM `df`",
        )
        worker = algorithm_factory.worker()
        worker.initialise(datasource=datasource, pod_identifier="user/df")
        assert worker.table == "df"
        results = worker.run()
        assert np.isclose(results.MAX_OF_G[0], 0.9997870068530033, atol=1e-4)

    @unit_test
    def test_worker_run_no_table_error(
        self, caplog: LogCaptureFixture, multi_table_excel_source: ExcelSource
    ) -> None:
        """Test that a SQL query returns correct result."""
        algorithm_factory = SqlQuery(
            query="SELECT MAX(G) AS MAX_OF_G FROM `df`",
        )
        worker = algorithm_factory.worker()
        worker.initialise(datasource=multi_table_excel_source, pod_identifier="user/df")
        worker.run()
        err_msg = "No table specified on which to execute the query on. "
        assert err_msg in caplog.text

    @unit_test
    def test_modeller_gets_sql_results(self, datasource: DataFrameSource) -> None:
        """Test that a SQL query returns a result."""
        algorithm_factory = SqlQuery(
            query="SELECT MAX(G) AS MAX_OF_G FROM `df`", table="df"
        )
        modeller = algorithm_factory.modeller()
        data = {"MAX_OF_G": [0.9997870068530033]}
        results = {"pod1": pd.DataFrame(data)}
        returned_results = modeller.run(results=results)
        assert np.isclose(
            returned_results["pod1"]["MAX_OF_G"], results["pod1"]["MAX_OF_G"], atol=1e-4
        )

    @integration_test
    def test_sql_algorithm_db_connection_multitable(
        self,
        db_session: sqlalchemy.engine.base.Engine,
    ) -> None:
        """Test sql algorithm on multitable db connection."""
        db_conn = DatabaseConnection(
            db_session, table_names=["dummy_data", "dummy_data_2"]
        )
        ds = DatabaseSource(db_conn, seed=420)
        algorithm_factory = SqlQuery(
            query='SELECT MAX("A") AS MAX_OF_A FROM dummy_data', table="dummy_data"
        )
        worker = algorithm_factory.worker()
        worker.datasource = ds
        res = worker.run()
        assert res["max_of_a"] is not None

    @unit_test
    def test_sql_output_duplicate_cols_error(
        self,
        caplog: LogCaptureFixture,
        datasource: DataFrameSource,
        mocker: MockerFixture,
    ) -> None:
        """Test that an error is raised if query output has a duplicated column name."""
        algorithm_factory = SqlQuery(query="SELECT * FROM `df`", table="df")
        worker = algorithm_factory.worker()
        dataset = pd.DataFrame({"A": [1], "B": [2]})
        dataset.columns = ["A", "A"]
        worker.datasource = datasource
        mocker.patch("pandasql.sqldf", return_value=dataset)
        worker.run()
        err_msg = "The following column names are duplicated in the output"
        assert err_msg in caplog.text

    @unit_test
    def test_sql_execute(
        self, mock_bitfount_session: Mock, mocker: MockerFixture
    ) -> None:
        """Test execute syntactic sugar."""
        query = SqlQuery(query="SELECT * FROM `df`")
        pod_identifiers = ["username/pod-id"]

        mock_modeller_run_method = mocker.patch.object(_Modeller, "run")
        query.execute(pod_identifiers=pod_identifiers)
        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers=pod_identifiers, require_all_pods=False
        )

    @unit_test
    def test_sql_algorithm_intermine(self, mocker: MockerFixture) -> None:
        """Test sql algorithm on Intermine data source."""
        service_url = "https://fake_url"

        mock_service = mocker.patch(
            "bitfount.data.datasources.intermine_source.Service"
        )
        ds = IntermineSource(service_url, token=None)
        mocker.patch.object(ds, "get_data", return_value=create_dataset())
        algorithm_factory = SqlQuery(
            query="SELECT * FROM `Gene_Protein`", table="Gene_Protein"
        )
        worker = algorithm_factory.worker()
        worker.datasource = ds
        res = worker.run()

        mock_service.assert_called_once_with(service_url, token=None)
        assert not res.empty


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for sql query algorithm."""

    def test_serialization(self) -> None:
        """Test Marshmallow Serialization for sql query algorithm."""
        algorithm_factory = SqlQuery(query="SELECT * FROM `df`")
        dumped = bf_dump(algorithm_factory)
        loaded = bf_load(dumped, _ALGORITHMS)
        assert isinstance(loaded, SqlQuery)
        assert algorithm_factory.__dict__ == loaded.__dict__


# Static tests for algorithm-protocol compatibility
if TYPE_CHECKING:
    from typing import cast

    from bitfount.federated.protocols.results_only import (
        _ResultsOnlyCompatibleAlgoFactory_,
        _ResultsOnlyCompatibleModellerAlgorithm,
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
    )

    # Check compatible with ResultsOnly
    _algo_factory: _ResultsOnlyCompatibleAlgoFactory_ = SqlQuery(
        query=cast(str, object()), pretrained_file=cast(str, object())
    )
    _modeller_side: _ResultsOnlyCompatibleModellerAlgorithm = _ModellerSide()
    _worker_side: _ResultsOnlyDataIncompatibleWorkerAlgorithm = _WorkerSide(
        query=cast(str, object()), table=cast(str, object())
    )



"""Tests databunch.py."""

import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.data.databunch import BitfountDataBunch
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import SchemaOverrideMapping
from tests.utils import PytestRequest
from tests.utils.helper import (
    TABLE_NAME,
    create_dataset,
    create_datasource,
    create_schema,
    unit_test,
)


@unit_test
class TestBitfountDataBunch:
    """Tests for BitfountDataBunch class."""

    # TODO: [BIT-1098] Unskip this test once different schemas are suppported
    @pytest.mark.skip(
        reason="This is testing a state that is not supported yet. "
        "i.e. the schema should not have columns that are missing in the data."
    )
    def test_create_datasets_with_missing_columns(self) -> None:
        """Test that columns are created if they exist in schema.

        This tests whether a BaseSource which is missing columns listed in the Schema
        has those missing columns added to it, to ensure that transformations
        can be applied across pods with varying schemas
        """
        dataset_with_all_columns = create_dataset()
        datasource_with_all_columns = DataFrameSource(dataset_with_all_columns)
        datasource_with_all_columns.load_data()
        schema_with_all_columns = BitfountSchema(
            datasource_with_all_columns,
            force_stypes={TABLE_NAME: {"categorical": ["TARGET"]}},
            table_name=TABLE_NAME,
        )
        dataset_missing_some_columns = dataset_with_all_columns.drop(
            columns=["A", "B", "I"]
        )
        datasource = DataFrameSource(dataset_missing_some_columns)
        databunch = BitfountDataBunch(
            data_structure=DataStructure(target="TARGET", table=TABLE_NAME),
            schema=schema_with_all_columns.tables[0],
            datasource=datasource,
        )

        assert databunch.validation_ds is not None
        assert databunch.test_ds is not None

        text_columns = ["I", "J", "K", "L"]
        assert set(databunch.train_ds.x_columns + databunch.train_ds.y_columns) == set(
            [i for i in dataset_with_all_columns.columns if i not in text_columns]
        )
        assert set(
            databunch.validation_ds.x_columns + databunch.validation_ds.y_columns
        ) == set([i for i in dataset_with_all_columns.columns if i not in text_columns])
        assert set(databunch.test_ds.x_columns + databunch.test_ds.y_columns) == set(
            [i for i in dataset_with_all_columns.columns if i not in text_columns]
        )


@unit_test
class TestCreateDataBunch:
    """Tests basic databunch generator."""

    @fixture(scope="function", params=[True, False])
    def datasource(self, request: PytestRequest) -> DataFrameSource:
        """Parameterised creation of DataFrameSource (with optional image col)."""
        data = create_dataset(image=request.param)
        if request.param:
            datasource = DataFrameSource(data, seed=420, image_col=["image"])
            datasource.load_data()
            return datasource

        datasource = DataFrameSource(data, seed=420)
        datasource.load_data()
        return datasource

    @fixture
    def datasource_w_loss_weights(self) -> BaseSource:
        """Creates a datasource with loss_weights column."""
        ds = create_datasource(classification=True, loss_weights=True)
        ds.load_data()
        return ds

    @fixture
    def schema(self) -> BitfountSchema:
        """Creates a schema."""
        return create_schema(classification=True, loss_weights=True)

    def test_databunch_local_with_schema(self, datasource: BaseSource) -> None:
        """Checks databunch creation with schema."""
        target = "TARGET"
        schema = BitfountSchema()
        schema.add_datasource_tables(
            datasource,
            force_stypes={TABLE_NAME: {"categorical": [target]}},
            table_name=TABLE_NAME,
        )
        datastructure = DataStructure(target=target, table=TABLE_NAME)

        db = BitfountDataBunch(
            data_structure=datastructure, schema=schema.tables[0], datasource=datasource
        )
        assert target == db.target

    def test_datastructure_selected_and_ignore_cols_updated(
        self, datasource_w_loss_weights: BaseSource, schema: BitfountSchema
    ) -> None:
        """Checks that the databunch updates the datstructure columns accordingly."""
        datastructure = DataStructure(
            target="TARGET",
            selected_cols=[
                "E",
                "F",
                "G",
                "H",
                "M",
                "N",
                "O",
                "P",
                "TARGET",
                "weights",
            ],
            loss_weights_col="weights",
            table=TABLE_NAME,
        )
        db = BitfountDataBunch(
            data_structure=datastructure,
            schema=schema.tables[0],
            datasource=datasource_w_loss_weights,
        )
        # Selected columns appear as specified in datastructure
        assert set(db.data_structure.selected_cols) == set(
            ["E", "F", "G", "H", "M", "N", "O", "P", "TARGET", "weights"]
        )
        # Any columns that are not part of the selected columns are ignored
        assert set(db.data_structure.ignore_cols) == set(
            ["A", "B", "C", "D", "I", "J", "K", "L", "Date"]
        )

    def test_databunch_text_columns_are_ignored_even_if_selected(
        self, datasource_w_loss_weights: BaseSource, schema: BitfountSchema
    ) -> None:
        """Checks that the text columns are ignored even if selected."""
        datastructure = DataStructure(
            target="TARGET",
            selected_cols=[
                "E",
                "F",
                "G",
                "H",
                "I",  # text
                "J",  # text
                "K",  # text
                "L",  # text
                "M",
                "N",
                "O",
                "P",
                "TARGET",
                "weights",
            ],
            loss_weights_col="weights",
            table=TABLE_NAME,
        )
        db = BitfountDataBunch(
            data_structure=datastructure,
            schema=schema.tables[0],
            datasource=datasource_w_loss_weights,
        )
        assert set(db.data_structure.ignore_cols) == set(
            ["A", "B", "C", "D", "I", "J", "K", "L", "Date"]
        )

    def test_databunch_with_datastructure_query(
        self, datasource: DataFrameSource, mocker: MockerFixture
    ) -> None:
        """Test databunch calls load_data with sql_query arg."""
        mock_load_data = mocker.patch.object(BaseSource, "load_data")
        mock_create_dataset = mocker.patch.object(BitfountDataBunch, "_create_datasets")

        target = "TARGET"
        schema = BitfountSchema()
        schema_override: SchemaOverrideMapping = {
            "categorical": [{"col1": {"value_1": 0, "value_2": 1}}]
        }
        schema.add_datasource_tables(
            datasource,
            force_stypes={TABLE_NAME: {"categorical": [target]}},
            table_name=TABLE_NAME,
        )
        datastructure = DataStructure(
            target=target,
            query="SELECT * from blah",
            schema_types_override=schema_override,
        )
        BitfountDataBunch(
            data_structure=datastructure, schema=schema.tables[0], datasource=datasource
        )
        mock_load_data.assert_called_once_with(sql_query="SELECT * from blah")
        mock_create_dataset.assert_called_once()



"""Utility functions for running modellers from configs."""
from __future__ import annotations

import asyncio
import logging
from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import desert
import yaml

from bitfount.data.datastructure import DataStructure
from bitfount.federated.authorisation_checkers import IdentityVerificationMethod
from bitfount.federated.exceptions import AggregatorError
from bitfount.federated.helper import (
    _check_and_update_pod_ids,
    _create_aggregator,
    _create_message_service,
)
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.federated.modeller import _Modeller
from bitfount.federated.privacy.differential import _DifferentiallyPrivate
from bitfount.federated.transport.config import MessageServiceConfig
from bitfount.federated.types import AlgorithmType, ProtocolType
from bitfount.federated.utils import _ALGORITHMS, _MODEL_STRUCTURES, _MODELS, _PROTOCOLS
from bitfount.hub.helper import _create_bitfounthub, get_pod_schema
from bitfount.models.base_models import _BaseModel
from bitfount.runners.config_schemas import (
    DataStructureConfig,
    ModellerConfig,
    TaskConfig,
)
from bitfount.types import DistributedModelProtocol

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub

DEFAULT_MODEL_OUT = Path("output-model.pt")

logger = logging.getLogger(__name__)


def setup_modeller_from_config_file(
    path_to_config_yaml: Union[str, PathLike],
) -> Tuple[_Modeller, List[str]]:
    """Creates a modeller from a YAML config file.

    Args:
        path_to_config_yaml: the path to the config file

    Returns:
        A tuple of the created Modeller and the list of pod identifiers to run
    """
    with open(path_to_config_yaml) as f:
        config_yaml = yaml.safe_load(f)
    config = desert.schema(ModellerConfig).load(config_yaml)
    return setup_modeller_from_config(config)


def setup_modeller_from_config(
    config: ModellerConfig,
) -> Tuple[_Modeller, List[str]]:
    """Creates a modeller from a loaded config mapping.

    Args:
        config: The modeller configuration.

    Returns:
        A tuple of the created Modeller and the list of pod identifiers to run
        the task against.
    """
    # Load config details
    transformation_file = config.task.transformation_file
    if transformation_file is not None and not transformation_file.exists():
        raise FileNotFoundError("Transformation file specified but doesn't exist")

    bitfount_hub = _create_bitfounthub(
        username=config.modeller.username, url=config.hub.url
    )

    # We assume that if the user has not included a username in
    # a pod identifier that it is their own pod
    pod_identifiers = _check_and_update_pod_ids(config.pods.identifiers, bitfount_hub)

    modeller = setup_modeller(
        pod_identifiers=pod_identifiers,
        data_config=config.data_structure,
        task_details=config.task,
        bitfount_hub=bitfount_hub,
        ms_config=config.message_service,
        identity_verification_method=config.modeller.identity_verification_method,
        private_key_file=config.modeller.private_key_file,
        idp_url=config.modeller._identity_provider_url,
    )

    return modeller, pod_identifiers


def setup_modeller(
    pod_identifiers: List[str],
    data_config: DataStructureConfig,
    task_details: TaskConfig,
    bitfount_hub: BitfountHub,
    ms_config: MessageServiceConfig,
    identity_verification_method: Union[
        str, IdentityVerificationMethod
    ] = IdentityVerificationMethod.DEFAULT,
    private_key_file: Optional[Path] = None,
    idp_url: Optional[str] = None,
) -> _Modeller:
    """Creates a modeller.

    Args:
        pod_identifiers: The pod identifiers of the pods to be used in the task.
        data_config: The details of the data schema and options to be used in the task.
        task_details: The task details as a TaskConfig instance.
        bitfount_hub: The BitfountHub instance.
        ms_config: The message service settings as a MessageServiceConfig instance.
        identity_verification_method: The identity verification method to use.
        private_key_file: The path to the private key used by this modeller.
        idp_url: URL of the modeller's identity provider.

    Returns:
        The created Modeller.
    """
    # Check validity of pod names
    if not pod_identifiers:
        raise ValueError("Must provide at least one `pod_identifier`")
    pod_identifiers = _check_and_update_pod_ids(pod_identifiers, bitfount_hub)

    # Check that the schemas of the given pods match
    # TODO: [BIT-1098] Manage pods with different schemas
    schema = get_pod_schema(pod_identifiers[0], hub=bitfount_hub)
    for pod_id in pod_identifiers[1:]:
        aux_schema = get_pod_schema(pod_id, hub=bitfount_hub)
        # We need to check that the schemas have the same contents
        if aux_schema != schema:
            raise ValueError(
                "Pod schemas must match in order to be able to train on them."
            )

    # Create data structure
    if not data_config.select.include and not data_config.select.exclude:
        data_config.select.include = schema.tables[0].get_feature_names()
    data_structure = DataStructure.create_datastructure(
        table_config=data_config.table_config,
        select=data_config.select,
        transform=data_config.transform,
        assign=data_config.assign,
    )

    # Create model
    model_details = task_details.model
    if model_details.structure:
        model_structure_class = _MODEL_STRUCTURES[model_details.structure.name]
        model_structure = model_structure_class(**model_details.structure.arguments)
    model: Union[_BaseModel, BitfountModelReference]
    if model_details.name:  # i.e. built-in model
        model_name = model_details.name
        try:
            model_class = _MODELS[model_name]
        except KeyError:
            raise KeyError(
                f"Unable to load built-in model {model_name}; "
                f"does this pod have the appropriate backend installed?"
            )
        # Check if Differential Privacy can be used with this model
        # The cast here is needed to assuage mypy due to the nested Schema classes
        # being different; any classes that inherit both _BaseModel and
        # DifferentiallyPrivate will override the schema anyway.
        if issubclass(
            cast(Type[_DifferentiallyPrivate], model_class), _DifferentiallyPrivate
        ):
            # Load defined model structure, if specified
            if model_details.structure:
                model = model_class(
                    datastructure=data_structure,
                    schema=schema,
                    model_structure=model_structure,
                    dp_config=model_details.dp_config,
                    logger_config=model_details.logger_config,
                    **model_details.hyperparameters,
                )

            else:
                model = model_class(
                    datastructure=data_structure,
                    schema=schema,
                    dp_config=model_details.dp_config,
                    logger_config=model_details.logger_config,
                    **model_details.hyperparameters,
                )

        else:
            # Load defined model structure, if specified
            if model_details.structure:
                model = model_class(
                    datastructure=data_structure,
                    schema=schema,
                    model_structure=model_structure,
                    logger_config=model_details.logger_config,
                    **model_details.hyperparameters,
                )

            else:
                model = model_class(
                    datastructure=data_structure,
                    schema=schema,
                    logger_config=model_details.logger_config,
                    **model_details.hyperparameters,
                )
    elif model_details.bitfount_model:  # i.e. custom model
        # Custom DP models not currently supported
        if model_details.dp_config:
            raise ValueError(
                "Custom models cannot currently be used with Differential Privacy."
            )

        # We set the hyperparameters of the BitfountModelReference using those from
        # the config; allows the config format to avoid duplicate hyperparameter
        # fields.
        model = BitfountModelReference(
            username=model_details.bitfount_model.username,
            model_ref=model_details.bitfount_model.model_ref,
            model_version=model_details.bitfount_model.model_version,
            datastructure=data_structure,
            schema=schema,
            hyperparameters=model_details.hyperparameters,
            hub=bitfount_hub,
        )

    else:
        raise TypeError(
            "Unrecognised model type: should be a built-in model "
            "or a BitfountModelReference."
        )

    # Set aggregation options
    if task_details.aggregator is not None:
        if model is not None:
            if not isinstance(
                model, (DistributedModelProtocol, BitfountModelReference)
            ):
                raise TypeError(
                    "Aggregation is only compatible with models implementing "
                    "DistributedModelProtocol or BitfountModelReference instances."
                )

        # We check early, whilst both are in scope, to ensure that, if weightings
        # have been supplied, weightings for all pods have been supplied.
        if task_details.aggregator.weights is not None:
            if (weight_pods := set(task_details.aggregator.weights.keys())) != (
                requested_pods := set(pod_identifiers)
            ):
                raise AggregatorError(
                    f"Pods in task and aggregation weightings do not match: "
                    f"{requested_pods} != {weight_pods}"
                )

        aggregator = _create_aggregator(
            secure_aggregation=task_details.aggregator.secure,
            weights=task_details.aggregator.weights,
        )
        task_details.protocol.arguments["aggregator"] = aggregator

    # Load algorithm from components
    algorithm = _ALGORITHMS[AlgorithmType(task_details.algorithm.name).name](
        **task_details.algorithm.arguments, model=model
    )

    # Load protocol from components
    protocol = _PROTOCOLS[ProtocolType(task_details.protocol.name).name](
        algorithm=algorithm,
        **task_details.protocol.arguments,
    )

    # Create Modeller
    message_service = _create_message_service(
        session=bitfount_hub.session,
        ms_config=ms_config,
    )
    modeller = _Modeller(
        protocol=protocol,
        message_service=message_service,
        bitfounthub=bitfount_hub,
        identity_verification_method=identity_verification_method,
        private_key=private_key_file,
        idp_url=idp_url,
    )

    return modeller


async def run_modeller_async(
    modeller: _Modeller,
    pod_identifiers: Iterable[str],
    require_all_pods: bool = False,
    model_out: Path = DEFAULT_MODEL_OUT,
) -> Optional[Any]:
    """Runs the modeller.

    Run the modeller, submitting tasks to the pods and waiting for the results.

    Args:
        modeller: The Modeller instance being used to manage the task.
        pod_identifiers: The group of pod identifiers to run the task against.
        require_all_pods: Require all pod identifiers specified to accept the task
            request to complete task execution.
        model_out: The path to save the model out to. Defaults to "./output-model.pt".

    Raises:
        PodResponseError: If require_all_pods is true and at least one pod
            identifier specifed rejects or fails to respond to a task request.
    """
    # Start task running
    result = await modeller.run_async(
        pod_identifiers, require_all_pods=require_all_pods
    )
    modeller._serialize(model_out)
    if result is False:
        return None

    return result


def run_modeller(
    modeller: _Modeller,
    pod_identifiers: Iterable[str],
    require_all_pods: bool = False,
    model_out: Path = DEFAULT_MODEL_OUT,
) -> Optional[Any]:
    """Runs the modeller.

    Run the modeller, submitting tasks to the pods and waiting for the results.

    Args:
        modeller: The Modeller instance being used to manage the task.
        pod_identifiers: The group of pod identifiers to run the task against.
        require_all_pods: Require all pod identifiers specified to accept the task
            request to complete task execution.
        model_out: The path to save the model out to. Defaults to "./output-model.pt".

    Raises:
        PodResponseError: If require_all_pods is true and at least one pod
            identifier specifed rejects or fails to respond to a task request.
    """
    pod_identifiers = _check_and_update_pod_ids(pod_identifiers, modeller._hub)
    return asyncio.run(
        run_modeller_async(modeller, pod_identifiers, require_all_pods, model_out)
    )



"""Modules for handling model data flow.

Data plugins can also be imported from this package.
"""
import importlib as _importlib
import inspect as _inspect
import logging as _logging
import pkgutil as _pkgutil
from typing import List as _List

from bitfount.config import BITFOUNT_PLUGIN_PATH as _BITFOUNT_PLUGIN_PATH
from bitfount.data import datasources as datasources
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasplitters import PercentageSplitter, SplitterDefinedInData
from bitfount.data.datastructure import DataStructure
from bitfount.data.exceptions import (
    BitfountSchemaError,
    DatabaseMissingTableError,
    DatabaseSchemaNotFoundError,
    DatabaseUnsupportedQueryError,
    DataNotLoadedError,
    DatasetSplitterError,
    DataStructureError,
    DuplicateColumnError,
)
from bitfount.data.helper import convert_epochs_to_steps
from bitfount.data.schema import BitfountSchema, TableSchema
from bitfount.data.types import (
    CategoricalRecord,
    ContinuousRecord,
    DataPathModifiers,
    DataSplit,
    ImageRecord,
    SemanticType,
    TextRecord,
)
from bitfount.data.utils import DatabaseConnection
from bitfount.utils import _get_module_from_file

_logger = _logging.getLogger(__name__)
__all__: _List[str] = [
    "BitfountSchema",
    "BitfountSchemaError",
    "CategoricalRecord",
    "ContinuousRecord",
    "DatabaseConnection",
    "DatabaseMissingTableError",
    "DatabaseSchemaNotFoundError",
    "DatabaseUnsupportedQueryError",
    "DatasetSplitterError",
    "DataNotLoadedError",
    "DataPathModifiers",
    "DataSplit",
    "DataStructure",
    "DataStructureError",
    "DuplicateColumnError",
    "ImageRecord",
    "PercentageSplitter",
    "SemanticType",
    "SplitterDefinedInData",
    "TableSchema",
    "TextRecord",
    "convert_epochs_to_steps",
]


# Import all concrete implementations of BaseSource in the datasources subdirectory
# as well as datasource plugins
for _module_info in _pkgutil.walk_packages(
    path=datasources.__path__ + [str(_BITFOUNT_PLUGIN_PATH / "datasources")],
):
    try:
        _module = _importlib.import_module(
            f"{datasources.__name__}.{_module_info.name}"
        )
    # Also catches `ModuleNotFoundError` which subclasses `ImportError`
    # Try to import the module from the plugin directory if it's not found in the
    # datasources directory
    except ImportError:
        try:
            _module = _get_module_from_file(
                _BITFOUNT_PLUGIN_PATH / "datasources" / f"{_module_info.name}.py",
            )
        except ImportError:
            _logger.debug(f"Error importing module {_module_info.name}")
    finally:
        for _, cls in _inspect.getmembers(_module, _inspect.isclass):
            if issubclass(cls, BaseSource) and not _inspect.isabstract(cls):
                globals().update({cls.__name__: getattr(_module, cls.__name__)})
                __all__.append(cls.__name__)
            # There are too many false positives if we don't restrict classes to those
            # that inherit from BaseSource for it to be a useful log message
            elif issubclass(cls, BaseSource) and cls.__name__ not in (
                "BaseSource",
                "MultiTableSource",
            ):
                _logger.debug(
                    f"Found class {cls.__name__} in module {_module_info.name} which "
                    f"did not fully implement BaseSource. Skipping."
                )


# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Federated logging."""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from bitfount.federated.transport.base_transport import _BaseMailbox


# Dynamic base class error below ignored as this is the recommended way to subclass
# logging.Logger as specified in the documentation:
# https://docs.python.org/3/library/logging.html#logging.getLoggerClass
class _FederatedLogger(logging.getLoggerClass()):  # type: ignore[misc] # Reason: see comment # noqa: B950
    """Federated Logger with extra federated methods."""

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def federated_debug(self, msg: object, *args: Any, **kwargs: Any) -> None:
        """Calls `self.debug` under the hood and sets the `federated` attribute."""
        self.debug(msg, *args, **kwargs, extra={"federated": True})

    def federated_info(self, msg: object, *args: Any, **kwargs: Any) -> None:
        """Calls `self.info` under the hood and sets the `federated` attribute."""
        self.info(msg, *args, **kwargs, extra={"federated": True})

    def federated_warning(self, msg: object, *args: Any, **kwargs: Any) -> None:
        """Calls `self.warning` under the hood and sets the `federated` attribute."""
        self.warning(msg, *args, **kwargs, extra={"federated": True})

    def federated_error(self, msg: object, *args: Any, **kwargs: Any) -> None:
        """Calls `self.error` under the hood and sets the `federated` attribute."""
        self.error(msg, *args, **kwargs, extra={"federated": True})

    def federated_critical(self, msg: object, *args: Any, **kwargs: Any) -> None:
        """Calls `self.critical` under the hood and sets the `federated` attribute."""
        self.critical(msg, *args, **kwargs, extra={"federated": True})


class _MailboxHandler(logging.Handler):
    """A class which sends records to a GRPC mailbox.

    Args:
        mailbox (BaseMailbox): the recipient of emitted logs
    """

    def __init__(self, mailbox: _BaseMailbox):
        super().__init__()
        self.mailbox = mailbox

    def emit(self, record: logging.LogRecord) -> None:
        """Emits a record to the designated mailbox.

        Args:
            record (logging.LogRecord): the record to be emitted
        """
        try:
            # Because emit() is overriding an inherited method, it cannot be async
            # def itself. However, the call to BaseMailbox _has_ to be, due to the
            # underlying async gRPC interaction. So we have to wrap this in
            # create_task() to make it async.
            # We should have an event loop by this point, as this is only accessed at
            # runtime and we have the main event loop running everything else.
            # If we don't, Handler.handleError() will deal with the raised exception.
            logging_task = asyncio.create_task(
                self.mailbox.log(record.__dict__), name="Federated Log Sending"
            )

            if record.levelno >= logging.ERROR:
                # Block flow until high priority messages have completed.
                asyncio.get_running_loop().run_until_complete(logging_task)
        except Exception:
            self.handleError(record)


class _FederatedLogFilter(logging.Filter):
    """Filters logs for Federated Logging."""

    def filter(self, record: logging.LogRecord) -> bool:
        """Only logs with the `federated` attribute set to True are passed through.

        Args:
            record (logging.LogRecord): the record to be filtered

        Returns:
            bool: whether or not the record passses through the filter to the handler(s)
        """
        if hasattr(record, "federated"):
            # Ignoring type error since we have already checked for the presence of
            # `federated` attribute
            if record.federated:  # type: ignore[attr-defined] # Reason: see comment # noqa: B950
                return True
        return False


def _federate_logger(mailbox: _BaseMailbox) -> None:
    """Adds a MailboxHandler to the `bitfount.federated` logger.

    Any existing mailbox handlers are first removed so that the handler and mailbox is
    updated prior to each task. This ensures that we don't attempt to log to a mailbox
    from an older task. The MailboxHandler is then attached with a filter which filters
    only for log messages that have the `federated` attribute set to True. We are
    federating the logger at the package level so that all loggers below it within the
    `bitfount.federated` package will also be federated.

    Args:
        mailbox (BaseMailbox): the mailbox used to create the mailbox handler
    """
    logger = _get_federated_logger("bitfount.federated")

    for handler in logger.handlers:
        if isinstance(handler, _MailboxHandler):
            logger.removeHandler(handler)

    mailbox_handler = _MailboxHandler(mailbox)
    mailbox_handler.setLevel(logging.DEBUG)
    log_filter = _FederatedLogFilter()
    mailbox_handler.addFilter(log_filter)
    logger.addHandler(mailbox_handler)


def _get_federated_logger(name: str) -> _FederatedLogger:
    """Returns logger with federated logging methods attached.

    This function should be used to retrieve the module level logger in any module in
    the `bitfount.federated` package where we want to send federated logs. We reset the
    logger class at the end of the function to ensure that subsequent calls to
    `logging.getLogger()` in other parts of the code don't return a federated logger.

    Returns:
        FederatedLogger: a regular logger but with extra methods for federated logging.
            It is important to note that these federated methods will not send messages
            in a federated way unless `federate_logger` has been called already.

    Raises:
        ValueError: if a federated logger is requested for a module/subpackage outside
            the `bitfount.federated` package
    """
    if not name.startswith("bitfount.federated"):
        raise ValueError("This can only be used from the federated package.")

    logging_class = logging.getLoggerClass()

    # Use the global logging lock for thread safety
    logging._acquireLock()  # type: ignore[attr-defined] # Reason: known private attribute, see comment # noqa: B950
    try:
        logging.setLoggerClass(_FederatedLogger)
        logger = logging.getLogger(name)
        if not isinstance(logger, _FederatedLogger):
            raise TypeError(
                f"logger for {name} is not federated; did a logger with this name"
                f" already exist?"
            )
        return logger
    finally:
        logging.setLoggerClass(logging_class)
        logging._releaseLock()  # type: ignore[attr-defined] # Reason: known private attribute # noqa: B950



"""Tests for base model algorithms module."""

from typing import Optional
from unittest.mock import Mock

import pytest

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseWorkerModelAlgorithm,
)
from bitfount.types import _SerializedWeights
from tests.utils.helper import create_datasource, unit_test


@unit_test
class TestBase:
    """Test base model algorithms module."""

    @pytest.fixture
    def datasource(self) -> BaseSource:
        """Fixture for datasource."""
        return create_datasource(classification=True)

    @pytest.mark.parametrize("model_params", [{"param": [1]}, None])
    def test_base_worker_initialise(
        self, datasource: BaseSource, model_params: Optional[_SerializedWeights]
    ) -> None:
        """Test initialise updates model parameters on worker side."""
        model = Mock(create_autospec=True)
        base = _BaseWorkerModelAlgorithm(model=model)
        base.initialise(datasource, model_params=model_params)
        if model_params:
            base.model.update_params.assert_called_once()  # type: ignore[attr-defined]  # Reason: model is a mock  # noqa: B950
        else:
            base.model.update_params.assert_not_called()  # type: ignore[attr-defined]  # Reason: model is a mock  # noqa: B950



"""Fixtures related to BitfountSchema."""
from unittest.mock import Mock, NonCallableMock, create_autospec

from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.data.schema import BitfountSchema


@fixture
def mock_bitfount_schema_load(mocker: MockerFixture) -> Mock:
    """Mocks out the BitfountSchema.load() classmethod.

    Mock will simply return the dict passed into the load() method.
    """
    mock_load: Mock = mocker.patch.object(BitfountSchema, "load", autospec=True)
    mock_load.side_effect = lambda data: data
    return mock_load


@fixture
def mock_bitfount_schema() -> NonCallableMock:
    """Mocked BitfountSchema for tests."""
    mock_schema: NonCallableMock = create_autospec(BitfountSchema, instance=True)
    return mock_schema



"""Message Service configuration."""
from dataclasses import dataclass
import logging

from grpc import aio, ssl_channel_credentials

from bitfount.federated.transport import _MESSAGE_SERVICE_GRPC_OPTIONS
from bitfount.federated.transport.protos.messages_pb2_grpc import MessageServiceStub

logger = logging.getLogger(__name__)

#: Production message service URL.
PRODUCTION_MESSAGE_SERVICE_URL = "messaging.bitfount.com"
_STAGING_MESSAGE_SERVICE_URL = "messaging.staging.bitfount.com"
_DEV_MESSAGE_SERVICE_URL = "localhost"
_DEV_MESSAGE_SERVICE_PORT = 5001
_DEV_MESSAGE_SERVICE_TLS = False


@dataclass
class MessageServiceConfig:
    """Configuration for the message service.

    Args:
        url: The URL of the message service. Defaults to
            `PRODUCTION_MESSAGE_SERVICE_URL`.
        port: The port of the message service. Defaults to 443.
        tls: Whether to use TLS. Defaults to True.
        use_local_storage: Whether to use local storage instead of communicating via the
            message service if both parties are on the same device. This can be used to
            remove the overhead of communication. Defaults to False.

    Raises:
        ValueError: If `tls` is False and `url` is a Bitfount URL.
    """

    url: str = PRODUCTION_MESSAGE_SERVICE_URL
    port: int = 443
    tls: bool = True  # only used for development
    use_local_storage: bool = False

    def __post_init__(self) -> None:
        if not self.tls and ".bitfount.com" in self.url:
            raise ValueError(
                "TLS disabled. Message service communication must be with TLS."
            )
        elif not self.tls:
            logger.warning("Message service communication without TLS.")

    # @cached_property can't be used because it returns the _exact_ same Awaitable
    # which cannot be awaited more than once.
    @property
    async def stub(self) -> MessageServiceStub:
        """Creates and returns message service stub from this config."""
        # This method must be kept as async def to ensure that the channel
        # creation occurs on the correct event loop.
        if self.tls:
            # This is an async secure_channel
            channel = aio.secure_channel(
                f"{self.url}:{self.port}",
                ssl_channel_credentials(),
                options=_MESSAGE_SERVICE_GRPC_OPTIONS,
            )
        else:
            # This is an async insecure_channel
            channel = aio.insecure_channel(
                f"{self.url}:{self.port}", options=_MESSAGE_SERVICE_GRPC_OPTIONS
            )

        return MessageServiceStub(channel)
